# MIT license
#
# Copyright (C) 2015 by XESS Corporation / Hildo Guillardi Junior
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

# Inserted by Pasteurize tool.
from __future__ import print_function
from __future__ import unicode_literals
from __future__ import division
from __future__ import absolute_import
from builtins import zip
from builtins import range
from builtins import int
from builtins import str
from future import standard_library
standard_library.install_aliases()

import future

import re, difflib
from bs4 import BeautifulSoup
import http.client # For web scraping exceptions.
from .. import urlencode, urlquote, urlsplit, urlunsplit
from .. import fake_browser
from .. import EXTRA_INFO_DIST, extra_info_dist_name_translations
from ...globals import PartHtmlError
from ...globals import logger, DEBUG_OVERVIEW, DEBUG_DETAILED, DEBUG_OBSESSIVE

from .. import distributor_dict
import pycountry

def define_locale_currency(locale_iso=None, currency_iso=None):
    '''@brief Configure the distributor for the country and currency intended.
    
    Scrape the configuration page and define the base URL of DigiKey for the
    currency and locale chosen.
    The currency is predominant over the locale/country and the defauld are
    currency='USD' and locale='US' for DigiKey.
    
    @param locale_iso `str` Country in ISO3166 alpha 2 standard.
    @param currency_iso `str` Currency in ISO4217 alpha 3 standard.'''
    url = 'https://www.digikey.com/en/resources/international'
    
    try:
        html = fake_browser(url, 4)
    except: # Could not get a good read from the website.
        logger.log(DEBUG_OBSESSIVE,'No HTML page for DigiKey configuration.')
        raise PartHtmlError
    html = BeautifulSoup(html, 'lxml')
    try:
        if currency_iso and not locale_iso:
            money = pycountry.currencies.get(alpha_3=currency_iso.upper())
            locale_iso = pycountry.countries.get(numeric=money.numeric).alpha_2
        if locale_iso:
            locale_iso = locale_iso.upper()
            country = pycountry.countries.get(alpha_2=locale_iso.upper()).name
            html = html.find('li', text=re.compile(country, re.IGNORECASE))
            url = html.find('a', id='linkcolor').get('href')
            
            distributor_dict['digikey']['site']['url'] = url
            distributor_dict['digikey']['site']['currency'] = pycountry.currencies.get(numeric=country.numeric).alpha_3
            distributor_dict['digikey']['site']['locale'] = locale_iso
    except:
        logger.log(DEBUG_OVERVIEW, 'Kept the last configuration {}, {} on {}.'.format(
                pycountry.currencies.get(alpha_3=distributor_dict['digikey']['site']['currency']).name,
                pycountry.countries.get(alpha_2=distributor_dict['digikey']['site']['locale']).name,
                distributor_dict['digikey']['site']['url']
            )) # Keep the current configuration.
    return


def get_extra_info(html_tree):
    '''@brief Get the extra characteristics `EXTRA_INFO_DIST` from the part web page.
       @param html_tree `str()` html of the distributor part page.
       @return `dict()` keys as characteristics names.
    '''
    info = {}
    try:
        table =  html_tree.find('table', id='prod-att-table')
        for row in table.find_all('tr', id=None): # `None`to ignore the header row.
            try:
                k = row.find('th').text.strip().lower()
                v = row.find('td').text.strip()
                k = extra_info_dist_name_translations.get(k, k)
                if k in EXTRA_INFO_DIST:
                    info[k] = v
            except:
                continue
        if 'datasheet' in EXTRA_INFO_DIST:
            try:
                info['datasheet'] = html_tree.find('a', href=True, target='_blank').get('href')
                if info['datasheet'][0:2]=='//':
                    info['datasheet'] = 'https:' + info['datasheet'] # Digikey missing definitions.
            except:
                pass
        if 'image' in EXTRA_INFO_DIST:
            try:
                info['image'] = html_tree.find('img', itemprop="image").get('src')
                if info['image'][0:2]=='//':
                    info['image'] = 'https:' + info['image'] # Digikey missing definitions.
            except:
                pass
    except AttributeError:
        # This happens when no pricing info is found in the tree.
        logger.log(DEBUG_OBSESSIVE, 'No Digikey pricing information found!')
    return info


def get_price_tiers(html_tree):
    '''@brief Get the pricing tiers from the parsed tree of the Digikey product page.
       @param html_tree `str()` html of the distributor part page.
       @return `dict()` price breaks, the keys are the quantities breaks.
    '''
    price_tiers = {}
    try:
        for tr in html_tree.find('table', id='product-dollars').find_all('tr'):
            try:
                td = tr.find_all('td')
                qty = int(re.sub('[^0-9]', '', td[0].text))
                price_tiers[qty] = float(re.sub('[^0-9\.]', '', td[1].text))
            except (TypeError, AttributeError, ValueError,
                    IndexError):  # Happens when there's no <td> in table row.
                continue
    except AttributeError:
        # This happens when no pricing info is found in the tree.
        logger.log(DEBUG_OBSESSIVE, 'No Digikey pricing information found!')
    return price_tiers


def part_is_reeled(html_tree):
    '''@brief Returns True if this Digi-Key part is reeled or Digi-reeled.
       @param html_tree `str()` html of the distributor part page.
       @return `True` or `False`.
    '''
    qty_tiers = list(get_price_tiers(html_tree).keys())
    if len(qty_tiers) > 0 and min(qty_tiers) >= 100:
        return True
    if html_tree.find('table',
                      id='product-details-reel-pricing') is not None:
        return True
    return False


def get_part_num(html_tree):
    '''@brief Get the part number from the Digikey product page.
       @param html_tree `str()` html of the distributor part page.
       @return `list()`of the parts that match.
    '''
    try:
        return re.sub('\s', '', html_tree.find('td',
                                               id='reportPartNumber').text)
    except AttributeError:
        logger.log(DEBUG_OBSESSIVE, 'No Digikey part number found!')
        return ''


def get_qty_avail(html_tree):
    '''@brief Get the available quantity of the part from the Digikey product page.
       @param html_tree `str()` html of the distributor part page.
       @return `int` avaliable quantity.
    '''
    try:
        qty_tree = html_tree.find('td', id='quantityAvailable').find('span', id='dkQty')
        qty_str = qty_tree.text
    except AttributeError:
        # No quantity found (not even 0) so this is probably a non-stocked part.
        # Return None so the part won't show in the spreadsheet for this dist.
        return None
    try:
        qty_str = re.search('([0-9,]*)', qty_str, re.IGNORECASE).group(1)
        return int(re.sub('[^0-9]', '', qty_str))
    except (AttributeError, ValueError):
        # Didn't find the usual quantity text field. This might be one of those
        # input fields for requesting a quantity, so get the value from the
        # input field.
        try:
            logger.log(DEBUG_OBSESSIVE, 'No Digikey part quantity found!')
            return int(qty_tree.find('input', type='text').get('value'))
        except (AttributeError, ValueError):
            # Well, there's a quantityAvailable section in the website, but
            # it doesn't contain anything decipherable. Let's just assume it's 0.
            return 0


def get_part_html_tree(dist, pn, extra_search_terms='', url=None, descend=2, local_part_html=None, scrape_retries=2):
    '''@brief Find the Digikey HTML page for a part number and return the URL and parse tree.
       @param dist
       @param pn Part number `str()`.
       @param extra_search_terms
       @param url
       @param descend
       @param local_part_html
       @param scrape_retries `int` Quantity of retries in case of fail.
       @return (html `str()` of the page, url)
    '''

    def merge_price_tiers(main_tree, alt_tree):
        '''Merge the price tiers from the alternate-packaging tree into the main tree.'''
        try:
            insertion_point = main_tree.find('table', id='product-dollars').find('tr')
            for tr in alt_tree.find('table', id='product-dollars').find_all('tr'):
                insertion_point.insert_after(tr)
        except AttributeError:
            logger.log(DEBUG_OBSESSIVE, 'Problem merging price tiers for Digikey part {} with alternate packaging!'.format(pn))

    def merge_qty_avail(main_tree, alt_tree):
        '''Merge the quantities from the alternate-packaging tree into the main tree.'''
        try:
            main_qty = get_qty_avail(main_tree)
            alt_qty = get_qty_avail(alt_tree)
            if main_qty is None:
                merged_qty = alt_qty
            elif alt_qty is None:
                merged_qty = main_qty
            else:
                merged_qty = max(main_qty, alt_qty)
            if merged_qty is not None:
                insertion_point = main_tree.find('td', id='quantityAvailable').find('span', id='dkQty')
                insertion_point.string = '{}'.format(merged_qty)
        except AttributeError:
            logger.log(DEBUG_OBSESSIVE, 'Problem merging available quantities for Digikey part {} with alternate packaging!'.format(pn))

    # Use the part number to lookup the part using the site search function, unless a starting url was given.
    if url is None:
        url = distributor_dict['digikey']['site']['url'] + '/products/en?keywords=' + urlquote(
        #'/scripts/DkSearch/dksus.dll?WT.z_header=search_go&lang=en&keywords=' + urlquote(
            pn + ' ' + extra_search_terms,
            safe='')
        #url = distributor_dict['digikey']['site']['url'] + '/product-search/en?KeyWords=' + urlquote(pn,safe='') + '&WT.z_header=search_go'
    elif url[0] == '/':
        url = distributor_dict['digikey']['site']['url'] + url

    # Open the URL, read the HTML from it, and parse it into a tree structure.
    try:
        html = fake_browser(url, scrape_retries)
    except:
        logger.log(DEBUG_OBSESSIVE,'No HTML page for {} from {}'.format(pn, dist))
        raise PartHtmlError

    # Abort if the part number isn't in the HTML somewhere.
    # (Only use the numbers and letters to compare PN to HTML.)
    if re.sub('[\W_]','',str.lower(pn)) not in re.sub('[\W_]','',str.lower(str(html))):
        logger.log(DEBUG_OBSESSIVE,'No part number {} in HTML page from {}'.format(pn, dist))
        raise PartHtmlError

    # Use the following code if Javascript challenge pages are used to block scrapers.
    # try:
    # ghst = Ghost()
    # sess = ghst.start(plugins_enabled=False, download_images=False, show_scrollbars=False, javascript_enabled=False)
    # html, resources = sess.open(url)
    # print('type of HTML is {}'.format(type(html.content)))
    # html = html.content
    # except Exception as e:
    # print('Exception reading with Ghost: {}'.format(e))

    try:
        tree = BeautifulSoup(html, 'lxml')
    except Exception:
        logger.log(DEBUG_OBSESSIVE,'No HTML tree for {} from {}'.format(pn, dist))
        raise PartHtmlError

    # If the tree contains the tag for a product page, then return it.
    if tree.find('div', class_='product-top-section') is not None:

        # Digikey separprint(ates cut-tape and reel packaging, so we need to examine more pages
        # to get all the pricing info. But don't descend any further if limit has been reached.
        if descend > 0:
            try:
                # Find all the URLs to alternate-packaging pages for this part.
                ap_urls = [
                    ap.find('li', class_='lnkAltPack').find_all('a')[-1].get('href')
                    for ap in tree.find(
                        'div', class_='bota',
                        id='additionalPackaging').find_all(
                            'ul', class_='more-expander-item')
                ]
                logger.log(DEBUG_OBSESSIVE,'Found {} alternate packagings for {} from {}'.format(len(ap_urls), pn, dist))
                ap_trees_and_urls = []  # Initialize as empty in case no alternate packagings are found.
                try:
                    ap_trees_and_urls = [get_part_html_tree(dist, pn, 
                                     extra_search_terms, ap_url, descend=0, scrape_retries=scrape_retries)
                                     for ap_url in ap_urls]
                except Exception:
                    logger.log(DEBUG_OBSESSIVE,'Failed to find alternate packagings for {} from {}'.format(pn, dist))

                # Put the main tree on the list as well and then look through
                # the entire list for one that's non-reeled. Use this as the
                # main page for the part.
                ap_trees_and_urls.append((tree, url))
                if part_is_reeled(tree):
                    for ap_tree, ap_url in ap_trees_and_urls:
                        if not part_is_reeled(ap_tree):
                            # Found a non-reeled part, so use it as the main page.
                            tree = ap_tree
                            url = ap_url
                            break  # Done looking.

                # Now go through the other pages, merging their pricing and quantity
                # info into the main page.
                for ap_tree, ap_url in ap_trees_and_urls:
                    if ap_tree is tree:
                        continue  # Skip examining the main tree. It already contains its info.
                    try:
                        # Merge the pricing info from that into the main parse tree to make
                        # a single, unified set of price tiers...
                        merge_price_tiers(tree, ap_tree)
                        # and merge available quantity, using the maximum found.
                        merge_qty_avail(tree, ap_tree)
                    except AttributeError:
                        logger.log(DEBUG_OBSESSIVE,'Problem merging price/qty for {} from {}'.format(pn, dist))
                        continue
            except AttributeError as e:
                logger.log(DEBUG_OBSESSIVE,'Problem parsing URLs from product page for {} from {}'.format(pn, dist))

        return tree, url  # Return the parse tree and the URL where it came from.

    # If the tree is for a list of products, then examine the links to try to find the part number.
    if tree.find('table', id='productTable') is not None:
        logger.log(DEBUG_OBSESSIVE,'Found product table for {} from {}'.format(pn, dist))
        if descend <= 0:
            logger.log(DEBUG_OBSESSIVE,'Passed descent limit for {} from {}'.format(pn, dist))
            raise PartHtmlError
        else:
            # Look for the table of products.
            products = tree.find(
                'table',
                id='productTable').find('tbody').find_all('tr')

            # Extract the product links for the part numbers from the table.
            # Extract links for both manufacturer and catalog numbers.
            product_links = [p.find('td',
                                    class_='tr-mfgPartNumber').a
                             for p in products]
            product_links.extend([p.find('td',
                                    class_='tr-dkPartNumber').a
                             for p in products])

            # Extract all the part numbers from the text portion of the links.
            part_numbers = [l.text for l in product_links]

            # Look for the part number in the list that most closely matches the requested part number.
            match = difflib.get_close_matches(pn, part_numbers, 1, 0.0)[0]

            # Now look for the link that goes with the closest matching part number.
            for l in product_links:
                if l.text == match:
                    # Get the tree for the linked-to page and return that.
                    logger.log(DEBUG_OBSESSIVE,'Selecting {} from product table for {} from {}'.format(l.text.strip(), pn, dist))
                    return get_part_html_tree(dist, pn, extra_search_terms,
                                              url=l.get('href', ''),
                                              descend=descend - 1, 
                                              scrape_retries=scrape_retries)

    # If the HTML contains a list of part categories, then give up.
    if tree.find('form', id='keywordSearchForm') is not None:
        logger.log(DEBUG_OBSESSIVE,'Found high-level part categories for {} from {}'.format(pn, dist))
        raise PartHtmlError

    # I don't know what happened here, so give up.
    logger.log(DEBUG_OBSESSIVE,'Unknown error for {} from {}'.format(pn, dist))
    raise PartHtmlError



# -*- coding: utf-8 -*-
# MIT license
#
# Copyright (C) 2018 by XESS Corporation / Hildo G Jr
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

# Author information.
__author__ = 'Hildo Guillardi Junior'
__webpage__ = 'https://github.com/hildogjr/'
__company__ = 'University of Campinas - Brazil'

# The global dictionary of supported EDA tools starts out empty.
eda_tool_dict = {}

import os

# The EDA tool directories will be found in this directory.
directory = os.path.dirname(__file__)

# Search for the EDA tool modules and import them.
eda_modules = {}
for module in os.listdir(directory):

    # Avoid importing non-directories.
    abs_module = os.path.join(directory, module)
    if not os.path.isdir(abs_module):
        continue

    # Avoid directories like __pycache__.
    if module.startswith('__'):
        continue

    # Import the module.
    eda_modules[module] = __import__(module, globals(), locals(), [], level=1)



# MIT license
#
# Copyright (C) 2018 by XESS Corporation / Hildo G Jr
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

# Author information.
__author__ = 'Hildo Guillardi Junior'
__webpage__ = 'https://github.com/hildogjr/'
__company__ = 'University of Campinas - Brazil'

# Libraries.
import sys
from bs4 import BeautifulSoup # XML file interpreter.
import multiprocessing # To deal with the parallel scrape.
import logging
from time import time
from random import choice
from ..eda_tools.eda_tools import order_refs # To better print the warnings about the parts.

try:
    # This is for Python 3.
    from urllib.parse import urlsplit, urlunsplit
except ImportError:
    # This is for Python 2.
    from urlparse import urlsplit, urlunsplit

from ..globals import logger, DEBUG_OVERVIEW, DEBUG_DETAILED, DEBUG_OBSESSIVE # Debug configurations.
from ..globals import SEPRTR
from ..globals import PartHtmlError
from . import distributor_dict

import os, re

# The distributor module directories will be found in this directory.
directory = os.path.dirname(__file__)

# Search for the distributor modules and import them.
dist_modules = {}
for module in os.listdir(directory):

    # Avoid importing non-directories.
    abs_module = os.path.join(directory, module)
    if not os.path.isdir(abs_module):
        continue

    # Avoid directories like __pycache__.
    if module.startswith('__'):
        continue

    # Import the module.
    dist_modules[module] = __import__(module, globals(), locals(), [], level=1)

__all__ = ['scrape_part', 'config_distributor']

def config_distributor(dist_name, locale_currency='USD'):
    '''@brief Configure the distributor for some locale/country and
    currency second ISO3166 and ISO4217
    
    @param `str` dist Distributor to configure.
    @param `str` Alpha 2 country or alpha 3 currency or even one slash other.'''
    try:
        dist_module = dist_modules[dist_name]
    except KeyError: # When use local distributor with personalized name.
        dist_module = dist_modules[distributor_dict[dist_name]['module']]
    try:
        if distributor_dict[dist_name]['scrape']=='web':
            # Not make sense to configurate a local distributor (yet).
            locale_currency = re.findall('\w{2,}', locale_currency)
            locale = None
            currency = None
            for alpha in locale_currency:
                if len(alpha)==2:
                    locale = alpha
                elif len(alpha)==3:
                    currency = alpha
            dist_module.define_locale_currency(locale_iso=locale, currency_iso=currency)
    except AttributeError:
        logger.warning('No currency/country configuration for {}.'.format(distributor_dict[dist_name]['label']))
        pass


def get_part_html_tree(part, dist, get_html_tree_func, local_part_html, scrape_retries, logger):
    '''@brief Get the HTML tree for a part.
    
    Get the HTML tree for a part from the given distributor website or local HTML.
    @param `str` part Part manufactor code or distributor stock code.
    @param `str` dist Distributor do scrape.
    @param `str` get_html_tree_func
    @param `str` local_part_html
    @param `int` scrape_retries Maximum times of web ritries.
    @param logger Logger handle.
    @return `str` with the HTML webpage.'''

    logger.log(DEBUG_OBSESSIVE, 'Looking in %s by %s:', distributor_dict[dist]['label'], order_refs(part.refs, True))

    for extra_search_terms in set([part.fields.get('manf', ''), '']):
        try:
            # Search for part information using one of the following:
            #    1) the distributor's catalog number.
            #    2) the manufacturer's part number.
            for key in (dist+'#', dist+SEPRTR+'cat#', 'manf#'):
                if key in part.fields:
                    if part.fields[key]:
                        # Founded manufacturer / distributor code valid (not empty).
                        return get_html_tree_func(dist, part.fields[key], extra_search_terms, local_part_html=local_part_html, scrape_retries=scrape_retries)
            # No distributor or manufacturer number, so give up.
            else:
                logger.warning("No '%s#' or 'manf#' field: cannot lookup part %s at %s.", dist, part.refs, dist)
                return BeautifulSoup('<html></html>', 'lxml'), ''
                #raise PartHtmlError
        except PartHtmlError:
            pass
        except AttributeError:
            break
    logger.warning("Part %s not found at %s.", order_refs(part.refs, False), distributor_dict[dist]['label'])
    # If no HTML page was found, then return a tree for an empty page.
    return BeautifulSoup('<html></html>', 'lxml'), ''


def scrape_part(args):
    '''@brief Scrape the data for a part from each distributor website or local HTML.
    
    Use distributors submodules to scrape each distributor part page and get
    informations such as price, quantity avaliable and others;
    
    @param `int` Count of the main loop.
    @param `str`String with the part number / distributor stock.
    @param `dict`
    @param `str`
    @param `int`Number of scrape retries.
    @param logger.getEffectiveLevel()
    @param throttle_lock
    @param throttle_tim
    @return id, url, `str` distributor stock part number, `dict` price tiers, `int` qty avail, `dict` extrainfo dist
    '''

    id, part, distributor_dict, local_part_html, scrape_retries, log_level, throttle_lock, throttle_timeouts = args # Unpack the arguments.

    if multiprocessing.current_process().name == "MainProcess":
        scrape_logger = logging.getLogger('kicost')
    else:
        scrape_logger = multiprocessing.get_logger()
        handler = logging.StreamHandler(sys.stdout)
        handler.setLevel(log_level)
        scrape_logger.addHandler(handler)
        scrape_logger.setLevel(log_level)

    # Create dictionaries for the various items of part data from each distributor.
    url = {}
    part_num = {}
    price_tiers = {}
    qty_avail = {}
    info_dist = {}

    # Scrape the part data from each distributor website or the local HTML.
    # Create a list of the distributor keys and randomly choose one of the
    # keys to scrape. After scraping, remove the distributor key.
    # Do this until all the distributors have been scraped.
    distributors = list(distributor_dict.keys())
    while distributors:

        d = choice(distributors)  # Randomly choose one of the available distributors.

        try:
            #dist_module = getattr(THIS_MODULE, d)
            dist_module = dist_modules[d]
        except KeyError: # When use local distributor with personalized name.
            dist_module = dist_modules[distributor_dict[d]['module']]

        # Try to access the list of distributor throttling timeouts.
        # Abort if some other process is already using the timeouts.
        if throttle_lock.acquire(blocking=False):

            # Check the throttling timeout for the chosen distributor to see if
            # another access to its website is allowed.
            if throttle_timeouts[d] <= time():

                # Update the timeout for this distributor website and release the sync. lock.
                throttle_timeouts[d] = time() + distributor_dict[d]['throttling_delay']
                throttle_lock.release()

                # Get the HTML tree for the part.
                html_tree, url[d] = get_part_html_tree(part, d, dist_module.get_part_html_tree, local_part_html, scrape_retries, scrape_logger)

                # Call the functions that extract the data from the HTML tree.
                part_num[d] = dist_module.get_part_num(html_tree)
                qty_avail[d] = dist_module.get_qty_avail(html_tree)
                price_tiers[d] = dist_module.get_price_tiers(html_tree)
                
                try:
                    # Get extra characeristics of the part in the web page.
                    # This will be use to comment in the 'cat#' column of the
                    # spreadsheet and some validations (in the future implementaions)
                    info_dist[d] = dist_module.get_extra_info(html_tree)
                except:
                    info_dist[d] = {}
                    pass

                # The part data has been scraped from this distributor, so remove it from the list.
                distributors.remove(d)

            # If the timeout for this distributor has not expired, then release
            # the sync. lock and try another distributor.
            else:
                throttle_lock.release()

    # Return the part data.
    return id, url, part_num, price_tiers, qty_avail, info_dist



# -*- coding: utf-8 -*-
# MIT license
#
# Copyright (C) 2018 by XESS Corporation / Hildo Guillardi Júnior
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

# Author information.
__author__ = 'Hildo Guillardi Junior'
__webpage__ = 'https://github.com/hildogjr/'
__company__ = 'University of Campinas - Brazil'

from .altium import get_part_groups

# Place information about this EDA into the eda_tool dictionary.
from .. import eda_tool_dict
eda_tool_dict.update(
    {
        'altium': {
            'module': 'altium', # The directory name containing this file.
            'label': 'Altium file', # Label used on the GUI.
            'desc': 'Altium Limited (formerly known as Protel until 2001).',
            # Formatting file match .
            'file': {
                'extension': '.xml', # File extension.
                'content': '\<GRID[\s\S]+<COLUMNS>[\s\S]+<COLUMN[\s\S]+<\/COLUMNS>[\s\S]+<ROWS>[\s\S]+\<ROW[\s\S]+\<\/ROWS>[\s\S]+\<\/GRID>' # Regular expression content match.
            }
        }
    }
)



# MIT license
#
# Copyright (C) 2018 by XESS Corporation / Hildo Guillardi Junior
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

# Inserted by Pasteurize tool.
from __future__ import print_function
from __future__ import unicode_literals
from __future__ import division
from __future__ import absolute_import
from builtins import zip, range, int, str
from future import standard_library
standard_library.install_aliases()

import future

import re, difflib
from bs4 import BeautifulSoup
from yattag import Doc, indent # For generating HTML page for local parts.
import copy # To be possible create more than one local distributor.
from .. import urlsplit, urlunsplit
from ...globals import PartHtmlError
from ...globals import logger, DEBUG_OVERVIEW, DEBUG_DETAILED, DEBUG_OBSESSIVE
from ...globals import SEPRTR


def create_part_html(parts, distributors):
    '''@brief Create HTML page containing info for local (non-webscraped) parts.
    @param parts `list()` of parts.
    @parm `list()`of the distributors to check each one is local.
    @return `str()` of the HTML page to be read by `get_part_html_tree()`
    '''
    
    logger.log(DEBUG_OVERVIEW, 'Create HTML page for parts with custom pricing...')
    
    doc, tag, text = Doc().tagtext()
    with tag('html'):
        with tag('body'):
            for p in parts:
                # Find the manufacturer's part number if it exists.
                pn = p.fields.get('manf#') # Returns None if no manf# field.

                # Find the various distributors for this part by
                # looking for leading fields terminated by SEPRTR.
                for key in p.fields:
                    try:
                        dist = key[:key.index(SEPRTR)]
                    except ValueError:
                        continue

                    # If the distributor is not in the list of web-scrapable distributors,
                    # then it's a local distributor. Copy the local distributor template
                    # and add it to the table of distributors.
                    if dist not in distributors:
                        distributors[dist] = copy.copy(distributors['local_template'])
                        distributors[dist]['label'] = dist  # Set dist name for spreadsheet header.

                # Now look for catalog number, price list and webpage link for this part.
                for dist in distributors:
                    cat_num = p.fields.get(dist+':cat#')
                    pricing = p.fields.get(dist+':pricing')
                    link = p.fields.get(dist+':link')
                    if cat_num is None and pricing is None and link is None:
                        continue

                    def make_random_catalog_number(p):
                        hash_fields = {k: p.fields[k] for k in p.fields}
                        hash_fields['dist'] = dist
                        return '#{0:08X}'.format(abs(hash(tuple(sorted(hash_fields.items())))))

                    cat_num = cat_num or pn or make_random_catalog_number(p)
                    p.fields[dist+':cat#'] = cat_num # Store generated cat#.
                    with tag('div', klass=dist+SEPRTR+cat_num):
                        with tag('div', klass='cat#'):
                            text(cat_num)
                        if pricing is not None:
                            with tag('div', klass='pricing'):
                                text(pricing)
                        if link is not None:
                            url_parts = list(urlsplit(link))
                            if url_parts[0] == '':
                                url_parts[0] = u'http'
                            link = urlunsplit(url_parts)
                            with tag('div', klass='link'):
                                text(link)

    # Remove the local distributor template so it won't be processed later on.
    # It has served its purpose.
    try:
        del distributors['local_template']
    except:
        pass

    html = doc.getvalue()
    if logger.isEnabledFor(DEBUG_OBSESSIVE):
        print(indent(html))
    return html


def get_price_tiers(html_tree):
    '''@brief Get the pricing tiers from the parsed tree of the local product page.
       @param html_tree `str()` html of the distributor part page.
       @return `dict()` price breaks, the keys are the quantities breaks.
    '''
    price_tiers = {}
    try:
        pricing = html_tree.find('div', class_='pricing').text
        pricing = re.sub('[^0-9.;:]', '', pricing) # Keep only digits, decimals, delimiters.
        for qty_price in pricing.split(';'):
            qty, price = qty_price.split(SEPRTR)
            price_tiers[int(qty)] = float(price)
    except AttributeError:
        # This happens when no pricing info is found in the tree.
        logger.log(DEBUG_OBSESSIVE, 'No local pricing information found!')
        return price_tiers  # Return empty price tiers.
    return price_tiers


def get_part_num(html_tree):
    '''@brief Get the part number from the local product page.
       @param html_tree `str()` html of the distributor part page.
       @return `list()`of the parts that match.
    '''
    try:
        part_num_str = html_tree.find('div', class_='cat#').text
        return part_num_str
    except AttributeError:
        return ''


def get_qty_avail(html_tree):
    '''@brief Get the available quantity of the part from the local product page.
       @param html_tree `str()` html of the distributor part page.
       @return `int` avaliable quantity.
    '''
    try:
        qty_str = html_tree.find('div', class_='quantity').text
    except (AttributeError, ValueError):
        # Return 0 (not None) so this part will show in the spreadsheet
        # even if there is no quantity found.
        return 0
    try:
        return int(re.sub('[^0-9]', '', qty_str))
    except ValueError:
        # Return 0 (not None) so this part will show in the spreadsheet
        # even if there is no quantity found.
        logger.log(DEBUG_OBSESSIVE, 'No local part quantity found!')
        return 0


def get_part_html_tree(dist, pn, extra_search_terms='', url=None, descend=None, local_part_html=None, scrape_retries=2):
    '''Extract the HTML tree from the HTML page for local parts.
       @param dist
       @param pn Part number `str()`.
       @param extra_search_terms
       @param url
       @param descend
       @param local_part_html
       @param scrape_retries `int` Quantity of retries in case of fail.
       @return (html `str()` of the page, `None`) The second argument is always `None` bacause there is not url to return.
    '''

    # Extract the HTML tree from the local part HTML page.
    try:
        tree = BeautifulSoup(local_part_html, 'lxml')
    except Exception:
        raise PartHtmlError

    try:
        # Find the DIV in the tree for the given part and distributor.
        class_ = dist + SEPRTR + pn
        part_tree = tree.find('div', class_=class_)
        url_tree = part_tree.find('div', class_='link')
        try:
            # Return the part data tree and any URL associated with the part.
            return part_tree, url_tree.text.strip()
        except AttributeError:
            # Return part data tree and None if the URL is not found.
            return part_tree, None
    except AttributeError:
        # Return an error if the part_tree is not found.
        raise PartHtmlError



# MIT license
#
# Copyright (C) 2018 by XESS Corporation / Hildo Guillardi Júnior
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

# Inserted by Pasteurize tool.
from __future__ import print_function
from __future__ import unicode_literals
from __future__ import division
from __future__ import absolute_import
from builtins import zip, range, int, str
from future import standard_library
standard_library.install_aliases()

import future

import re, difflib
from bs4 import BeautifulSoup
import http.client # For web scraping exceptions.
from .. import urlencode, urlquote, urlsplit, urlunsplit
from .. import fake_browser
from ...globals import PartHtmlError
from ...globals import logger, DEBUG_OVERVIEW, DEBUG_DETAILED, DEBUG_OBSESSIVE
from currency_converter import CurrencyConverter
currency = CurrencyConverter()


def get_price_tiers(html_tree):
    '''@brief Get the pricing tiers from the parsed tree of the RS Components product page.
       @param html_tree `str()` html of the distributor part page.
       @return `dict()` price breaks, the keys are the quantities breaks.
    '''
    price_tiers = {}
    
    try:
        for row in html_tree.find_all('div', class_='table-row value-row'):
            qty = row.find('div',
                        class_='breakRangeWithoutUnit col-xs-4').text
            price = row.find('div',
                        class_='unitPrice col-xs-4').text
            try:
                qty = int( re.findall('\s*([0-9\,]+)', qty)[0] )
                price = re.sub('[^0-9\.]', '', price.replace(',','.') )
                price = currency.convert(float(price), 'EUR', 'USD')
                price_tiers[qty] = price
            except (TypeError, AttributeError, ValueError):
                continue
    except AttributeError:
        # This happens when no pricing info is found in the tree.
        return price_tiers  # Return empty price tiers.
    return price_tiers
    
def get_part_num(html_tree):
    '''@brief Get the part number from the RS product page.
       @param html_tree `str()` html of the distributor part page.
       @return `dict()` price breaks, the keys are the quantities breaks.
    '''
    try:
        pn_str = html_tree.find('span', class_='keyValue').text
        pn = re.sub('[^0-9\-]','', pn_str)
        return pn
    except KeyError:
        return '' # No catalog number found in page.
    except AttributeError:
        return '' # No ProductDescription found in page.

def get_qty_avail(html_tree):
    '''Get the available quantity of the part from the RS product page.
       @param html_tree `str()` html of the distributor part page.
       @return `int` avaliable quantity.
    '''
        
    try:
        # Note that 'availability' is misspelled in the container class name!        
        qty_str = html_tree.find('span', class_=('stock-msg-content', 'table-cell')).text
    except (AttributeError, ValueError):
        # No quantity found (not even 0) so this is probably a non-stocked part.
        # Return None so the part won't show in the spreadsheet for this dist.
        return None
    try:
        qty = re.sub('[^0-9]','',qty_str[0:10])  # Strip all non-number chars.
        return int(qty)  # Return integer for quantity.
    except ValueError:
        # No quantity found (not even 0) so this is probably a non-stocked part.
        # Return None so the part won't show in the spreadsheet for this dist.
        return None

def get_part_html_tree(dist, pn, extra_search_terms='', url=None, descend=2, local_part_html=None, scrape_retries=2):
    '''@brief Find the RS Components HTML page for a part number and return the URL and parse tree.
       @param dist
       @param pn Part number `str()`.
       @param extra_search_terms
       @param url
       @param descend
       @param local_part_html
       @param scrape_retries `int` Quantity of retries in case of fail.
       @return (html `str()` of the page, url)
    '''
            
    # Use the part number to lookup the part using the site search function, unless a starting url was given.
    if url is None:
        url = 'http://it.rs-online.com/web/c/?searchTerm=' + urlquote(pn + ' ' + extra_search_terms, safe='')

    elif url[0] == '/':
        url = 'http://it.rs-online.com' + url
    elif url.startswith('..'):
        url = 'http://it.rs-online.com/Search/' + url

    # Open the URL, read the HTML from it, and parse it into a tree structure.
    try:
        html = fake_browser(url, scrape_retries)
    except:
        logger.log(DEBUG_OBSESSIVE,'No HTML page for {} from {}'.format(pn, dist))
        raise PartHtmlError

    try:
        tree = BeautifulSoup(html, 'lxml')
    except Exception:
        logger.log(DEBUG_OBSESSIVE,'No HTML tree for {} from {}'.format(pn, dist))
        raise PartHtmlError

    # Abort if the part number isn't in the HTML somewhere.
    # (Only use the numbers and letters to compare PN to HTML.)
    if re.sub('[\W_]','',str.lower(pn)) not in re.sub('[\W_]','',str.lower(str(html))):
        logger.log(DEBUG_OBSESSIVE,'No part number {} in HTML page from {}'.format(pn, dist))
        raise PartHtmlError
        
    # If the tree contains the tag for a product page, then just return it.
    if tree.find('div', class_='advLineLevelContainer'):
        return tree, url

    # If the tree is for a list of products, then examine the links to try to find the part number.
    if tree.find('div', class_=('resultsTable','results-table-container')) is not None:
        logger.log(DEBUG_OBSESSIVE,'Found product table for {} from {}'.format(pn, dist))
        if descend <= 0:
            logger.log(DEBUG_OBSESSIVE,'Passed descent limit for {} from {}'.format(pn, dist))
            raise PartHtmlError
        else:
            # Look for the table of products.
            products = tree.find('table', id='results-table').find_all(
                    'tr', class_='resultRow')

            # Extract the product links for the part numbers from the table.
            product_links = [p.find('a', class_='product-name').get('href') for p in products]

            # Extract all the part numbers from the text portion of the links.
            part_numbers = [p.find('span', class_='text-contents').get_text() for p in products]

            # Look for the part number in the list that most closely matches the requested part number.
            match = difflib.get_close_matches(pn, part_numbers, 1, 0.0)[0]

            # Now look for the link that goes with the closest matching part number.
            for i in range(len(product_links)):
                if part_numbers[i] == match:
                    # Get the tree for the linked-to page and return that.
                    logger.log(DEBUG_OBSESSIVE,'Selecting {} from product table for {} from {}'.format(part_numbers[i], pn, dist))
                    return get_part_html_tree(dist, pn, extra_search_terms,
                                              url=product_links[i],
                                              descend=descend-1,
                                              scrape_retries=scrape_retries)

    # I don't know what happened here, so give up.
    logger.log(DEBUG_OBSESSIVE,'Unknown error for {} from {}'.format(pn, dist))
    raise PartHtmlError



# -*- coding: utf-8 -*-

__author__='Giacinto Luigi Cerone'

from .rs import *

# Place information about this distributor into the distributor dictionary.
from .. import distributor_dict
distributor_dict.update(
    {
        'rs': {
            'module': 'rs',           # The directory name containing this file.
            'scrape': 'web',          # Allowable values: 'web' or 'local'.
            'label': 'RS Components', # Distributor label used in spreadsheet columns.
            'order_cols': ['part_num', 'purch', 'refs'],  # Sort-order for online orders.
            'order_delimiter': ' ',  # Delimiter for online orders.
            # Formatting for distributor header in worksheet.
            'wrk_hdr_format': {
                'font_size': 14,
                'font_color': 'white',
                'bold': True,
                'align': 'center',
                'valign': 'vcenter',
                'bg_color': '#FF0000'  # RS Components red.
            },
            # Web site defitions.
            'site': {
                'url': 'http://rs-online.com/',
                'currency': 'USD',
                'locale': 'UK'
            },
        }
    }
)



# MIT license
#
# Copyright (C) 2018 by XESS Corporation / Hildo G Jr
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

# Inserted by Pasteurize tool.
from __future__ import print_function
from __future__ import unicode_literals
from __future__ import division
from __future__ import absolute_import
from builtins import open
from future import standard_library
standard_library.install_aliases()

import argparse as ap # Command argument parser.
import os, sys, platform
import logging, time
#import inspect # To get the internal module and informations of a module/class.
from .kicost import * # kicost core functions.
try:
    from .kicost_gui import * # User guide.
except ImportError:
    pass # If the wxPython dependences are not installed and
         # the user just want the KiCost CLI.
from .distributors import distributor_dict
from .eda_tools import eda_tool_dict
from . import __version__ # Version control by @xesscorp.

NUM_PROCESSES = 30  # Maximum number of parallel web-scraping processes.
HTML_RESPONSE_RETRIES = 2 # Number of attempts to retrieve part data from a website.

from .globals import *
logger = logging.getLogger('kicost')

###############################################################################
# Command-line interface.
###############################################################################

def main():

    parser = ap.ArgumentParser(
        description='Build cost spreadsheet for a KiCAD project.')
    parser.add_argument('-v', '--version',
                        action='version',
                        version='KiCost ' + __version__)
    parser.add_argument('-i', '--input',
                        nargs='+',
                        type=str,
                        metavar='FILE.XML',
                        help='One or more schematic BOM XML files.')
    parser.add_argument('-o', '--output',
                        nargs='?',
                        type=str,
                        metavar='FILE.XLSX',
                        help='Generated cost spreadsheet.')
    parser.add_argument('-f', '--fields',
                        nargs='+',
                        type=str,
                        default=[],
                        metavar='NAME',
                        help='''Specify the names of additional part fields to 
                            extract and insert in the global data section of 
                            the spreadsheet.''')
    parser.add_argument('-var', '--variant',
                        nargs='+',
                        type=str,
                        default=' ', # Default variant is a space.
                        help='schematic variant name filter.')
    parser.add_argument('-w', '--overwrite',
                        action='store_true',
                        help='Allow overwriting of an existing spreadsheet.')
    parser.add_argument('-s', '--serial',
                        action='store_true',
                        help='Do web scraping of part data using a single process.')
    parser.add_argument('-q', '--quiet',
                        action='store_true',
                        help='Enable quiet mode with no warnings.')
    parser.add_argument('-np', '--num_processes',
                        nargs='?',
                        type=int,
                        default=NUM_PROCESSES,
                        const=NUM_PROCESSES,
                        metavar='NUM_PROCESSES',
                        help='''Set the number of parallel 
                            processes used for web scraping part data.''')
    parser.add_argument('-ign', '--ignore_fields',
                        nargs='+',
                        default=[],
                        help='Declare part fields to ignore when reading the BoM file.',
                        metavar='NAME',
                        type=str)
    parser.add_argument('-grp', '--group_fields',
                        nargs='+',
                        default=[],
                        help='Declare part fields to merge when grouping parts.',
                        metavar='NAME',
                        type=str)
    parser.add_argument('-d', '--debug',
                        nargs='?',
                        type=int,
                        default=None,
                        metavar='LEVEL',
                        help='Print debugging info. (Larger LEVEL means more info.)')
    parser.add_argument('-eda', '--eda_tool', choices=['kicad', 'altium', 'csv'],
                        nargs='+',
                        default='kicad',
                        help='Choose EDA tool from which the XML BOM file originated, or use csv for .CSV files.')
    parser.add_argument('--show_dist_list',
                        action='store_true',
                        help='Show list of distributors that can be scraped for cost data, then exit.')
    parser.add_argument('--show_eda_list',
                        action='store_true',
                        help='Show list of EDA tools whose files KiCost can read, then exit.')
    parser.add_argument('--no_collapse',
                        action='store_true',
                        help='Do not collapse the part references in the spreadsheet.')
    parser.add_argument('-e', '--exclude',
                        nargs='+', type=str, default='',
                        metavar = 'DIST',
                        help='Excludes the given distributor(s) from the scraping process.')
    parser.add_argument('--include',
                        nargs='+', type=str, default='',
                        metavar = 'DIST',
                        help='Includes only the given distributor(s) in the scraping process.')
    parser.add_argument('--no_scrape',
                        action='store_true',
                        help='Create a spreadsheet without scraping part data from distributor websites.')
    parser.add_argument('-rt', '--retries',
                        nargs='?',
                        type=int,
                        default=HTML_RESPONSE_RETRIES,
                        metavar = 'NUM_RETRIES',
                        help='Specify the number of attempts to retrieve part data from a website.')
    parser.add_argument('--throttling_delay',
                        nargs='?', type=float, default=0.0,
                        metavar='DELAY',
                        help="Specify minimum delay (in seconds) between successive accesses to a distributor's website.")
    parser.add_argument('--currency', '--locale',
                        nargs='?',
                        type=str,
                        default='USD',
                        help='Define the priority locale/country and currency on the scrape. Use the ISO4217 for currency and ISO3166:2 for country. Input e.g.: `US`, `USD`, `US-USD` or `EUR-US`. Currency is priritized over the locale/country. If give country with more than one currency, it will be chosen, in the sequence, `USD`, `EUR` or alphabetical order. Default: `USD`.')
    parser.add_argument('--user',
                        action='store_true',
                        help='Start the user guide to run KiCost passing the file parameter give by "--input", all others parameters are ignored.')


    args = parser.parse_args()

    # Set up logging.
    if args.debug is not None:
        log_level = logging.DEBUG + 1 - args.debug
    elif args.quiet is True:
        log_level = logging.ERROR
    else:
        log_level = logging.WARNING
    #handler = logging.StreamHandler(sys.stdout)
    #handler.setLevel(log_level)
    #logger.addHandler(handler) # It's not necessary to add a handle here, the default is already `sys.stdout` and adding twice it creates the BUG #193, doesn't allowing to use correctly the `tqdm` (process bar) print handle.
    logger.setLevel(log_level)

    if args.show_dist_list:
        print('Distributor list:', *sorted(list(distributor_dict.keys())))
        return
    if args.show_eda_list:
        #eda_names = [o[0] for o in inspect.getmembers(eda_tools_imports) if inspect.ismodule(o[1])]
        #print('EDA supported list:', ', '.join(eda_names))
        print('EDA supported list:', *sorted(list(eda_tool_dict.keys())))
        return

    # Set up spreadsheet output file.
    if args.output == None:
        # If no output file is given...
        if args.input != None:
            # Send output to spreadsheet with name of input file.
            if len(args.input)>1:
                # Compose a name with the multiple BOM input file names.
                args.output = output_filename_multipleinputs(args.input)
            else:
                args.output = os.path.splitext(args.input[0])[0] + '.xlsx'
        else:
            # Send output to spreadsheet with name of this application.
            args.output = os.path.splitext(sys.argv[0])[0] + '.xlsx'
    else:
        # Output file was given. Make sure it has spreadsheet extension.
        args.output = os.path.splitext(args.output)[0] + '.xlsx'

    # Call the KiCost interface to alredy run KiCost, this is just to use the
    # saved user configurations of the graphical interface.
    if args.user:
        try:
            kicost_gui_run([os.path.abspath(fileName) for fileName in args.input])
        except (ImportError,NameError):
            kicost_gui_notdependences()
            #kicost_gui_run([os.path.abspath(fileName) for fileName in args.input])
        return

    # Handle case where output is going into an existing spreadsheet file.
    if os.path.isfile(args.output):
        if not args.overwrite:
            logger.critical('''Output file {} already exists! Use the
                --overwrite option to replace it.'''.format(args.output))
            sys.exit(1)

    # Set XML input source.
    if args.input == None:
        try:
            kicost_gui() # Use the user guide.
        except (ImportError,NameError):
            kicost_gui_notdependences()
            #kicost_gui()
        return
    else:
        # Otherwise get XML from the given file.
        for i in range(len(args.input)):
            # Set '.xml' as the default file extension, treating this exception
            # allow other files extension and formats.
            try:
                if os.path.splitext(args.input[i])[1] == '':
                    args.input[i] += '.xml'
                elif os.path.splitext(args.input[i])[1] == '.csv' or args.eda_tool[i] == 'csv':
                    args.eda_tool = 'csv'
            except IndexError:
                pass

    # Set number of processes to use for web scraping.
    if args.serial:
        num_processes = 1
    else:
        num_processes = args.num_processes

    # Remove all the distributor from the list for not scrape any web site.
    if args.no_scrape:
        dist_list = None
    else:
        if not args.include:
            dist_list = list(distributor_dict.keys())
        else:
            dist_list = args.include
        for d in args.exclude:
            dist_list.remove(d)

    logger.log(DEBUG_OBSESSIVE, 'Started KiCost v.{} on {}({}) Python {}.{}.{}'.format(
                                              __version__,
                                              platform.platform(),
                                              platform.architecture()[0],
                                              sys.version_info.major,
                                              sys.version_info.minor,
                                              sys.version_info.micro)
                                          )
    #try:
    kicost(in_file=args.input, eda_tool_name=args.eda_tool,
        out_filename=args.output, collapse_refs=not args.no_collapse,
        user_fields=args.fields, ignore_fields=args.ignore_fields,
        group_fields=args.group_fields, variant=args.variant,
        dist_list=dist_list, num_processes=num_processes,
        scrape_retries=args.retries, throttling_delay=args.throttling_delay,
        local_currency=args.currency)
    #except Exception as e:
    #    sys.exit(e)

###############################################################################
# Main entrypoint.
###############################################################################
if __name__ == '__main__':
    start_time = time.time()
    main()
    logger = logging.getLogger('kicost')
    logger.log(logging.DEBUG-2, 'Elapsed time: %f seconds', time.time() - start_time)


###############################################################################
# Additional functions
###############################################################################

def kicost_gui_notdependences():
    print('You don\'t have the wxPython dependence to run the GUI interface. Run once of the follow commands in terminal to install them:')
    print('pip3 install -U wxPython # For Windows & macOS')

    print('pip install -U -f https://extras.wxpython.org/wxPython4/extras/linux/gtk3/ubuntu-16.04 wxPython # For Linux 16.04')
    print('Or download from last version from <https://wxpython.org/pages/downloads/>')
    sys.exit(1)



# MIT license
#
# Copyright (C) 2015 by XESS Corporation
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

# Inserted by Pasteurize tool.
from __future__ import print_function
from __future__ import unicode_literals
from __future__ import division
from __future__ import absolute_import
from builtins import zip
from builtins import range
from builtins import int
from builtins import str
from future import standard_library
standard_library.install_aliases()

import future

import re
import difflib
import json
from bs4 import BeautifulSoup
import http.client # For web scraping exceptions.
from .. import urlencode, urlquote, urlsplit, urlunsplit
from .. import fake_browser
from ...globals import PartHtmlError
from ...globals import logger, DEBUG_OVERVIEW, DEBUG_DETAILED, DEBUG_OBSESSIVE

HTML_RESPONSE_RETRIES = 2

def __ajax_details(pn):
    '''@brief Load part details from TME using XMLHttpRequest.
       @param pn `str()` part number
       @return (html, quantity avaliable)
    '''
    data = urlencode({
        'symbol': pn,
        'currency': 'USD'
    }).encode("utf-8")
    
    try:
        html = fake_browser('https://www.tme.eu/en/_ajax/ProductInformationPage/_getStocks.html', 4, ('X-Requested-With', 'XMLHttpRequest') )
    except: # Couldn't get a good read from the website.
        logger.log(DEBUG_OBSESSIVE,'No AJAX data for {} from {}'.format(pn, 'TME'))
        return None, None

    try:
        r = r.decode('utf-8')  # Convert bytes to string in Python 3.
        p = json.loads(r).get('Products')
        if p is not None and isinstance(p, list):
            p = p[0]
            html_tree = BeautifulSoup(p.get('PriceTpl', '').replace("\n", ""), "lxml")
            quantity = p.get('InStock', '0')
            return html_tree, quantity
        else:
            return None, None
    except (ValueError, KeyError, IndexError):
        logger.log(DEBUG_OBSESSIVE, 'Could not obtain AJAX data from TME!')
        return None, None

def get_price_tiers(html_tree):
    '''@brief Get the pricing tiers from the parsed tree of the TME product page.
       @param html_tree `str()` html of the distributor part page.
       @return `dict()` price breaks, the keys are the quantities breaks.
    '''
    price_tiers = {}
    try:
        pn = get_part_num(html_tree)
        if pn == '':
            return price_tiers

        ajax_tree, quantity = __ajax_details(pn)
        if ajax_tree is None:
            return price_tiers

        qty_strs = []
        price_strs = []
        for tr in ajax_tree.find('tbody', id='prices_body').find_all('tr'):
            td = tr.find_all('td')
            if len(td) == 3:
                qty_strs.append(td[0].text)
                price_strs.append(td[2].text)

        qtys_prices = list(zip(qty_strs, price_strs))
        for qty_str, price_str in qtys_prices:
            try:
                qty = re.search('(\s*)([0-9,]+)', qty_str).group(2)
                qty = int(re.sub('[^0-9]', '', qty))
                price_tiers[qty] = float(re.sub('[^0-9\.]', '', price_str))
            except (TypeError, AttributeError, ValueError, IndexError):
                continue
    except AttributeError:
        # This happens when no pricing info is found in the tree.
        logger.log(DEBUG_OBSESSIVE, 'No TME pricing information found!')
        return price_tiers  # Return empty price tiers.
    return price_tiers


def get_part_num(html_tree):
    '''@brief Get the part number from the TME product page.
       @param html_tree `str()` html of the distributor part page.
       @return `list()`of the parts that match.
    '''
    try:
        return html_tree.find('td', class_="pip-product-symbol").text
    except AttributeError:
        logger.log(DEBUG_OBSESSIVE, 'No TME part number found!')
        return ''


def get_qty_avail(html_tree):
    '''@brief Get the available quantity of the part from the TME product page.
       @param html_tree `str()` html of the distributor part page.
       @return `int` avaliable quantity.
    '''
    pn = get_part_num(html_tree)
    if pn == '':
        logger.log(DEBUG_OBSESSIVE, 'No TME part quantity found!')
        return None

    ajax_tree, qty_str = __ajax_details(pn)
    if qty_str is None:
        return None

    try:
        return int(qty_str)
    except ValueError:
        # No quantity found (not even 0) so this is probably a non-stocked part.
        # Return None so the part won't show in the spreadsheet for this dist.
        logger.log(DEBUG_OBSESSIVE, 'No TME part quantity found!')
        return None


def get_part_html_tree(dist, pn, extra_search_terms='', url=None, descend=2, local_part_html=None, scrape_retries=2):
    '''@brief Find the TME HTML page for a part number and return the URL and parse tree.
       @param dist
       @param pn Part number `str()`.
       @param extra_search_terms
       @param url
       @param descend
       @param local_part_html
       @param scrape_retries `int` Quantity of retries in case of fail.
       @return (html `str()` of the page, url)
    '''

    global HTML_RESPONSE_RETRIES
    HTML_RESPONSE_RETRIES = scrape_retries

    # Use the part number to lookup the part using the site search function, unless a starting url was given.
    if url is None:
        url = 'https://www.tme.eu/en/katalog/?search=' + urlquote(
            pn + ' ' + extra_search_terms,
            safe='')
    elif url[0] == '/':
        url = 'https://www.tme.eu' + url

    # Open the URL, read the HTML from it, and parse it into a tree structure.
    try:
        html = fake_browser(url, scrape_retries)
    except:
        logger.log(DEBUG_OBSESSIVE,'No HTML page for {} from {}'.format(pn, dist))
        raise PartHtmlError

    # Abort if the part number isn't in the HTML somewhere.
    # (Only use the numbers and letters to compare PN to HTML.)
    if re.sub('[\W_]','',str.lower(pn)) not in re.sub('[\W_]','',str.lower(str(html))):
        logger.log(DEBUG_OBSESSIVE,'No part number {} in HTML page from {} ({})'.format(pn, dist, url))
        raise PartHtmlError

    try:
        tree = BeautifulSoup(html, 'lxml')
    except Exception:
        logger.log(DEBUG_OBSESSIVE,'No HTML tree for {} from {}'.format(pn, dist))
        raise PartHtmlError

    # If the tree contains the tag for a product page, then just return it.
    if tree.find('div', id='ph') is not None:
        return tree, url

    # If the tree is for a list of products, then examine the links to try to find the part number.
    if tree.find('table', id="products") is not None:
        logger.log(DEBUG_OBSESSIVE,'Found product table for {} from {}'.format(pn, dist))
        if descend <= 0:
            logger.log(DEBUG_OBSESSIVE,'Passed descent limit for {} from {}'.format(pn, dist))
            raise PartHtmlError
        else:
            # Look for the table of products.
            products = tree.find(
                'table',
                id="products").find_all(
                    'tr',
                    class_=('product-row'))

            # Extract the product links for the part numbers from the table.
            product_links = []
            for p in products:
                for a in p.find('td', class_='product').find_all('a'):
                    product_links.append(a)

            # Extract all the part numbers from the text portion of the links.
            part_numbers = [l.text for l in product_links]

            # Look for the part number in the list that most closely matches the requested part number.
            match = difflib.get_close_matches(pn, part_numbers, 1, 0.0)[0]

            # Now look for the link that goes with the closest matching part number.
            for l in product_links:
                try:
                    if (not l.get('href', '').startswith('./katalog')) and l.text == match:
                        # Get the tree for the linked-to page and return that.
                        logger.log(DEBUG_OBSESSIVE,'Selecting {} from product table for {} from {}'.format(l.text, pn, dist))
                        # TODO: The current implementation does up to four HTTP
                        # requests per part (search, part details page for TME P/N,
                        # XHR for pricing information, and XHR for stock
                        # availability). This is mainly for the compatibility with
                        # other distributor implementations (html_tree gets passed
                        # to all functions).
                        # A modified implementation (which would pass JSON data
                        # obtained by the XHR instead of the HTML DOM tree) might be
                        # able to do the same with just two requests (search for TME
                        # P/N, XHR for pricing and stock availability).
                        return get_part_html_tree(dist, pn, extra_search_terms,
                                                  url=l.get('href', ''),
                                                  descend=descend-1,
                                                  scrape_retries=scrape_retries)
                except KeyError:
                    pass    # This happens if there is no 'href' in the link, so just skip it.

    # I don't know what happened here, so give up.
    logger.log(DEBUG_OBSESSIVE,'Unknown error for {} from {}'.format(pn, dist))
    raise PartHtmlError



# -*- coding: utf-8 -*-
# MIT license
#
# Copyright (C) 2018 by XESS Corporation / Hildo Guillardi Junior
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

__author__ = 'XESS Corporation'
__email__ = 'info@xess.com'

from random import choice

import http.client # For web scraping exceptions.
try:
    # This is for Python 3
    from urllib.parse import urlencode, quote_plus as urlquote, urlsplit, urlunsplit
    from urllib.request import urlopen, Request
    import urllib.error
    WEB_SCRAPE_EXCEPTIONS = (urllib.error.URLError, http.client.HTTPException)
except ImportError:
    # This is for Python 2
    from urlparse import urlsplit, urlunsplit
    from urllib import urlencode, quote_plus as urlquote
    from urllib2 import urlopen, Request
    import urllib2
    WEB_SCRAPE_EXCEPTIONS = (urllib2.URLError, http.client.HTTPException)


def get_user_agent():
    ''' The default user_agent_list comprises chrome, IE, firefox, Mozilla, opera, netscape.
      You can find more user agent strings at https://techblog.willshouse.com/2012/01/03/most-common-user-agents/.
      Used for the function `fake_browser(url, retries)`
    '''
    user_agent_list = [
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/62.0.3202.94 Safari/537.36",
        "Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/62.0.3202.94 Safari/537.36",
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/61.0.3163.100 Safari/537.36",
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:57.0) Gecko/20100101 Firefox/57.0",
        "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_13_1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/62.0.3202.94 Safari/537.36",
        "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_13_1) AppleWebKit/604.3.5 (KHTML, like Gecko) Version/11.0.1 Safari/604.3.5",
        "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_12_6) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/62.0.3202.94 Safari/537.36",
        "Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/61.0.3163.100 Safari/537.36",
        "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_12_6) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/61.0.3163.100 Safari/537.36",
        "Mozilla/5.0 (Windows NT 6.1; Win64; x64; rv:57.0) Gecko/20100101 Firefox/57.0",
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:56.0) Gecko/20100101 Firefox/56.0",
        "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_12_6) AppleWebKit/604.3.5 (KHTML, like Gecko) Version/11.0.1 Safari/604.3.5",
        "Mozilla/5.0 (X11; Linux x86_64; rv:57.0) Gecko/20100101 Firefox/57.0",
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/62.0.3202.89 Safari/537.36",
        "Mozilla/5.0 (Windows NT 6.1; WOW64; Trident/7.0; rv:11.0) like Gecko",
        "Mozilla/5.0 (Windows NT 6.1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/62.0.3202.94 Safari/537.36",
        "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_11_6) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/62.0.3202.94 Safari/537.36",
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/52.0.2743.116 Safari/537.36 Edge/15.15063",
        "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/62.0.3202.94 Safari/537.36",
        "Mozilla/5.0 (Windows NT 6.3; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/62.0.3202.94 Safari/537.36",
        "Mozilla/5.0 (Windows NT 6.1; Win64; x64; rv:56.0) Gecko/20100101 Firefox/56.0",
        "Mozilla/5.0 (Macintosh; Intel Mac OS X 10.13; rv:57.0) Gecko/20100101 Firefox/57.0",
        "Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:56.0) Gecko/20100101 Firefox/56.0",
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/58.0.3029.110 Safari/537.36 Edge/16.16299",
        "Mozilla/5.0 (Macintosh; Intel Mac OS X 10.12; rv:57.0) Gecko/20100101 Firefox/57.0",
        "Mozilla/5.0 (Windows NT 6.1; Win64; x64; rv:52.0) Gecko/20100101 Firefox/52.0",
        "Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/62.0.3202.89 Safari/537.36",
        "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_11_6) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/61.0.3163.100 Safari/537.36",
        "Mozilla/5.0 (Windows NT 10.0; WOW64; Trident/7.0; rv:11.0) like Gecko",
        "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/62.0.3202.89 Safari/537.36",
        "Mozilla/5.0 (X11; Linux x86_64; rv:52.0) Gecko/20100101 Firefox/52.0",
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/62.0.3202.89 Safari/537.36 OPR/49.0.2725.47",
        "Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:57.0) Gecko/20100101 Firefox/57.0",
        "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_13_1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/61.0.3163.100 Safari/537.36",
        "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_11_6) AppleWebKit/604.3.5 (KHTML, like Gecko) Version/11.0.1 Safari/604.3.5",
        "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_12_6) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/62.0.3202.89 Safari/537.36",
        "Mozilla/5.0 (Macintosh; Intel Mac OS X 10.12; rv:56.0) Gecko/20100101 Firefox/56.0",
        "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_3) AppleWebKit/600.5.17 (KHTML, like Gecko) Version/8.0.5 Safari/600.5.17",
        "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_13) AppleWebKit/604.1.38 (KHTML, like Gecko) Version/11.0 Safari/604.1.38",
        "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/62.0.3202.75 Safari/537.36",
        "Mozilla/5.0 (Windows NT 6.3; Win64; x64; rv:57.0) Gecko/20100101 Firefox/57.0",
        "Mozilla/5.0 (Windows NT 6.3; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/61.0.3163.100 Safari/537.36",
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/62.0.3202.75 Safari/537.36",
        "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_12_6) AppleWebKit/604.1.38 (KHTML, like Gecko) Version/11.0 Safari/604.1.38",
        "Mozilla/5.0 (Windows NT 6.1; WOW64; rv:52.0) Gecko/20100101 Firefox/52.0",
        "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/61.0.3163.100 Safari/537.36",
        "Mozilla/5.0 (Macintosh; Intel Mac OS X 10.11; rv:57.0) Gecko/20100101 Firefox/57.0",
        "Mozilla/5.0 (Windows NT 10.0; WOW64; rv:56.0) Gecko/20100101 Firefox/56.0",
        "Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/62.0.3202.94 Safari/537.36",
        "Mozilla/5.0 (Windows NT 6.1; rv:57.0) Gecko/20100101 Firefox/57.0",
        "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_5) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/62.0.3202.94 Safari/537.36",
        "Mozilla/5.0 (Windows NT 6.3; Win64; x64; rv:56.0) Gecko/20100101 Firefox/56.0",
        "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_13_1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/62.0.3202.89 Safari/537.36",
        "Mozilla/5.0 (Windows NT 6.1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/61.0.3163.100 Safari/537.36",
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/62.0.3202.89 Safari/537.36 OPR/49.0.2725.39",
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:58.0) Gecko/20100101 Firefox/58.0",
        "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_5) AppleWebKit/603.3.8 (KHTML, like Gecko) Version/10.1.2 Safari/603.3.8",
        "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_12_6) AppleWebKit/603.3.8 (KHTML, like Gecko) Version/10.1.2 Safari/603.3.8",
        "Mozilla/5.0 (Windows NT 6.1; rv:52.0) Gecko/20100101 Firefox/52.0",
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/51.0.2704.79 Safari/537.36 Edge/14.14393",
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/61.0.3163.100 Safari/537.36 OPR/48.0.2685.52",
        "Mozilla/5.0 (Windows NT 10.0; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/62.0.3202.94 Safari/537.36",
        "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Ubuntu Chromium/62.0.3202.75 Chrome/62.0.3202.75 Safari/537.36",
        "Mozilla/5.0 (Windows NT 10.0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/62.0.3202.94 Safari/537.36",
        "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/51.0.2704.106 Safari/537.36",
        "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_13_0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/62.0.3202.94 Safari/537.36",
        "Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/62.0.3202.75 Safari/537.36",
        "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/62.0.3202.62 Safari/537.36",
        "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Ubuntu Chromium/62.0.3202.94 Chrome/62.0.3202.94 Safari/537.36",
        "Mozilla/5.0 (compatible; MSIE 9.0; Windows NT 6.1; Trident/5.0;  Trident/5.0)",
        "Mozilla/5.0 (Windows NT 10.0; WOW64; Trident/7.0; Touch; rv:11.0) like Gecko",
        "Mozilla/5.0 (X11; Linux x86_64; rv:56.0) Gecko/20100101 Firefox/56.0",
        "Mozilla/5.0 (X11; Linux x86_64; rv:58.0) Gecko/20100101 Firefox/58.0",
        "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_13_0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/61.0.3163.100 Safari/537.36",
        "Mozilla/5.0 (iPad; CPU OS 11_1_2 like Mac OS X) AppleWebKit/604.3.5 (KHTML, like Gecko) Version/11.0 Mobile/15B202 Safari/604.1",
        "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_12_6) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/62.0.3202.75 Safari/537.36",
        "Mozilla/5.0 (Windows NT 6.1; Trident/7.0; rv:11.0) like Gecko",
        "Mozilla/5.0 (compatible; MSIE 9.0; Windows NT 6.0; Trident/5.0;  Trident/5.0)",
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/60.0.3112.113 Safari/537.36",
        "Mozilla/5.0 (Windows NT 6.1; rv:56.0) Gecko/20100101 Firefox/56.0",
        "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_5) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/61.0.3163.100 Safari/537.36",
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/62.0.3202.62 Safari/537.36",
        "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_12_5) AppleWebKit/603.2.4 (KHTML, like Gecko) Version/10.1.1 Safari/603.2.4",
        "Mozilla/5.0 (Windows NT 10.0; WOW64; rv:52.0) Gecko/20100101 Firefox/52.0",
        "Mozilla/5.0 (Macintosh; Intel Mac OS X 10.13; rv:56.0) Gecko/20100101 Firefox/56.0",
        "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Ubuntu Chromium/62.0.3202.89 Chrome/62.0.3202.89 Safari/537.36",
        "Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/61.0.3163.100 Safari/537.36",
        "Mozilla/5.0 (Windows NT 6.2; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/62.0.3202.94 Safari/537.36",
        "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_11_6) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/62.0.3202.89 Safari/537.36",
        "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_12_5) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/61.0.3163.100 Safari/537.36",
        "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_12_5) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/62.0.3202.94 Safari/537.36",
        "Mozilla/5.0 (Windows NT 5.1; rv:52.0) Gecko/20100101 Firefox/52.0",
        "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/59.0.3071.115 Safari/537.36",
    ]
    return choice(user_agent_list)


# Open the URL, read the HTML from it, and parse it into a tree structure.
    
def fake_browser(url, scrape_retries=4, add_header=None):
    for _ in range(scrape_retries):
        try:
            req = Request(url)
            if add_header:
                req.add_header(add_header)
            req.add_header('Accept-Language', 'en-US')
            req.add_header('Accept', 'text/html')
            req.add_header('Cookie', 'foo=bar')
            req.add_header('User-agent', get_user_agent())
            response = urlopen(req)
            html = response.read()
            break
        except WEB_SCRAPE_EXCEPTIONS:
            logger.log(DEBUG_DETAILED,'Exception while web-scraping {}'.format(url))
            pass
    else:
        raise ValueError('No page')
    return html

# Extra informations to by got by each part in the distributors.
EXTRA_INFO_DIST = ['value', 'tolerance', 'footprint', 'power', 'current', 'voltage', 'frequency', 'temp_coeff', 'manf',
              'size', 'op temp', 'orientation', 'color',
              'datasheet', 'image', # Links.
             ]
extra_info_dist_name_translations = {
    #TODO it will need to put here language translation after implementation of ISSUE #65?
    'resistance': 'value',
    'inductance': 'value',
    'capacitance': 'value',
    'manufacturer': 'manf',
    'package': 'footprint',
    'package / case': 'footprint',
    'datasheets': 'datasheet',
    'dimension': 'size',
    'size / dimension': 'size',
    'operating temperature': 'op temp',
    'voltage - rated': 'voltage',
    'Mating Orientation': 'orientation',
    'coulor': 'color',
    'wire gauge': 'wire',
}

# The global dictionary of distributor information starts out empty.
distributor_dict = {}



# -*- coding: utf-8 -*- 
# MIT license
#
# Copyright (C) 2018 by XESS Corporation / Hildo Guillardi Junior
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

# Author information.
__author__ = 'Hildo Guillardi Junior'
__webpage__ = 'https://github.com/hildogjr/'
__company__ = 'University of Campinas - Brazil'

# Libraries.
try:
    import wx # wxWidgets for Python.
except ImportError:
    raise ImportError('wxPython package not recognised.')
import webbrowser # To update informations.
import sys, os, subprocess # To access OS commands and run in the shell.
import threading
import time # To elapse time.
import platform # To check the system platform when open the XLS file.
import tempfile # To create the temporary log file.
from datetime import datetime # To create the log name, when asked to save.
from distutils.version import StrictVersion # To comparasion of versions.
import re # Regular expression parser.

from . import __version__ # Version control by @xesscorp.
from .kicost import *  # kicost core functions.
from .distributors import distributor_dict, fake_browser # Use the configurations alredy made to get KiCost last version.
from .eda_tools import eda_tool_dict
from .eda_tools.eda_tools import file_eda_match

__all__ = ['kicost_gui', 'kicost_gui_run']

# Guide definitions.
FILE_HIST_QTY = 10
SEP_FILES = '\n' # File separator in the comboBox.
WILDCARD = "BoM compatible formats (*.xml,*.csv)|*.xml;*.csv|"\
            "KiCad/Altium BoM file (*.xml)|*.xml|" \
            "Proteus/Generic BoM file (*.csv)|*.csv"
CONFIG_FILE = 'KiCost' # Config file for Linux and Windows registry key for KiCost configurations.
GUI_SIZE_ENTRY = 'GUI_size'
GUI_POSITION_ENTRY = 'GUI_position'
PAGE_OFFICIAL = 'https://xesscorp.github.io/KiCost/'
PAGE_UPDATE = 'https://pypi.python.org/pypi/kicost' # Page with the last official version.
#https://github.com/xesscorp/KiCost/blob/master/kicost/version.py



#======================================================================
def open_file(filepath):
    '''@brief Open a file with the default application by yht different OS.
       @param filepath str() file name.
    '''
    if sys.platform.startswith('darwin'): # Mac-OS.
        subprocess.call(('open', filepath))
    elif sys.platform.startswith('windows'): # Windows.
        os.startfile(filepath)
    elif sys.platform.startswith('linux'): # Linux.
        subprocess.call(('xdg-open', filepath))
    else:
        print('Not recognized OS.')


#======================================================================
class FileDropTarget( wx.FileDropTarget ):
    ''' This object implements Drop Target functionality for Files.
        @param Window handle.
    '''
    def __init__(self, obj):
        ''' @brief Constructor.'''
        wx.FileDropTarget.__init__(self)
        self.obj = obj
    
    def OnDropFiles(self, x, y, filenames):
        #self.obj.SetInsertionPointEnd()
        self.obj.addFile(filenames)
        return True # No error.


#======================================================================
class menuDistributors( wx.Menu ):
    ''' @brief Menu of the istributor checkbox list. Provide select all, unselect and toggle hotkey.
        @param TextBox handle.
    '''
    def __init__( self, parent ):
        ''' @brief Constructor.'''
        super(menuDistributors, self).__init__()
        self.parent = parent
        
        mmi = wx.MenuItem(self, wx.NewId(), 'Select &all')
        self.Append(mmi)
        self.Bind(wx.EVT_MENU, self.selectAll, mmi)
        
        mmi = wx.MenuItem(self, wx.NewId(), '&Unselect all')
        self.Append(mmi)
        self.Bind(wx.EVT_MENU, self.unselectAll, mmi)
        
        mmi = wx.MenuItem(self, wx.NewId(), '&Toggle')
        self.Append(mmi)
        self.Bind(wx.EVT_MENU, self.toggleAll, mmi)
    
    def selectAll( self, event ):
        ''' @brief Select all distributor that exist.'''
        event.Skip()
        for idx in range(self.parent.m_checkList_dist.GetCount()):
            if not self.parent.m_checkList_dist.IsChecked(idx):
                self.parent.m_checkList_dist.Check(idx)
    
    def unselectAll( self, event ):
        ''' @brief Unselect all distributor that exist.'''
        event.Skip()
        for idx in range(self.parent.m_checkList_dist.GetCount()):
            if self.parent.m_checkList_dist.IsChecked(idx):
                self.parent.m_checkList_dist.Check(idx, False)
    
    def toggleAll( self, event ):
        ''' @brief Toggle all distributor that exist.'''
        event.Skip()
        for idx in range(self.parent.m_checkList_dist.GetCount()):
            if self.parent.m_checkList_dist.IsChecked(idx):
                self.parent.m_checkList_dist.Check(idx, False)
            else:
                self.parent.m_checkList_dist.Check(idx)


#======================================================================
class menuMessages( wx.Menu ):
    ''' @brief Menu of the messages text. Provide copy and save options.
        @param TextBox handle.
    '''
    def __init__( self, parent ):
        ''' @brief Constructor.'''
        super(menuMessages, self).__init__()
        self.parent = parent
        
        mmi = wx.MenuItem(self, wx.NewId(), '&Purge')
        self.Append(mmi)
        self.Bind(wx.EVT_MENU, self.purgeMessages, mmi)
        
        self.AppendSeparator()
        
        mmi = wx.MenuItem(self, wx.NewId(), '&Copy to clipboard')
        self.Append(mmi)
        self.Bind(wx.EVT_MENU, self.copyMessages, mmi)
        
        mmi = wx.MenuItem(self, wx.NewId(), 'Cut to clipboard')
        self.Append(mmi)
        self.Bind(wx.EVT_MENU, self.cutMessages, mmi)
        
        mmi = wx.MenuItem(self, wx.NewId(), '&Save')
        self.Append(mmi)
        self.Bind(wx.EVT_MENU, self.saveMessages, mmi)
        
        mmi = wx.MenuItem(self, wx.NewId(), 'S&ave and clear')
        self.Append(mmi)
        self.Bind(wx.EVT_MENU, self.saveClearMessages, mmi)
        
        mmi = wx.MenuItem(self, wx.NewId(), '&Open externally')
        self.Append(mmi)
        self.Bind(wx.EVT_MENU, self.openMessages, mmi)
        
    
    def copyMessages( self, event ):
        ''' @brief Copy the warning/error/log messages to clipboard.'''
        event.Skip()
        clipdata = wx.TextDataObject()
        clipdata.SetText( self.parent.m_textCtrl_messages.GetValue() )
        wx.TheClipboard.Open()
        wx.TheClipboard.SetData(clipdata)
        wx.TheClipboard.Close()
    
    def cutMessages( self, event ):
        ''' @brief Cut the warning/error/log messages to clipboard.'''
        event.Skip()
        self.purgeMessages(event)
        self.copyMessages(event)
    
    def purgeMessages( self, event ):
        ''' @brief Clear message box.'''
        event.Skip()
        self.parent.m_textCtrl_messages.Clear()
    
    def saveMessages( self, event ):
        ''' @brief Save the messages as a text "KiCost*.log" file.'''
        event.Skip()
        actualDir = (os.getcwd() if self.parent.m_comboBox_files.GetValue() else \
            os.path.dirname(os.path.abspath( self.parent.m_comboBox_files.GetValue() )) )
        dlg = wx.FileDialog(
            self.parent, message = "Save log as...",
            defaultDir = actualDir, 
            defaultFile = "KiCost " + datetime.now().strftime('%Y-%m-%d %Hh%Mmin%Ss'),
            wildcard = "Log file (*.log)|*.log",
            style = wx.FD_SAVE|wx.FD_OVERWRITE_PROMPT
            )
        dlg.SetFilterIndex(0)
        if dlg.ShowModal() == wx.ID_OK:
            f = open( dlg.GetPath() , 'w')
            f.write( self.parent.m_textCtrl_messages.GetValue() )
            f.close()
            wx.MessageBox('The log file as saved.', 'Info', wx.OK | wx.ICON_INFORMATION)
        dlg.Destroy()
    
    def saveClearMessages( self, event ):
        '''@brief Save the messages and clear the log in the guide.'''
        event.Skip()
        self.saveMessages(event)
        self.purgeMessages(event)
    
    def openMessages( self, event ):
        ''' @brief Save the messages in a temporary file and open it in the default text editor before sytem deletation.'''
        event.Skip()
        #TODO - not working on Ubuntu
        #self.parent.m_textCtrl_messages.AppendText('\naqui\nhjhk')
        with tempfile.NamedTemporaryFile(prefix='KiCost_', suffix='.log', delete=True, mode='w') as temp:
            temp.write( self.parent.m_textCtrl_messages.GetValue() )
            open_file(temp.name)
            temp.close()


#======================================================================
class formKiCost ( wx.Frame ):
    ''' @brief Main frame / form of KiCost GUI.'''
    
    def __init__( self, parent ):
        ''' @brief Constructor, code generated by wxFormBuilder tool.'''
        #### **  Begin of the guide code generated by wxFormBulilder software, available in <https://github.com/wxFormBuilder/wxFormBuilder/>  ** ####
        wx.Frame.__init__ ( self, parent, id = wx.ID_ANY, title = u"KiCost", pos = wx.DefaultPosition, size = wx.Size( 446,351 ), style = wx.DEFAULT_FRAME_STYLE|wx.TAB_TRAVERSAL )
        
        self.SetSizeHints( wx.DefaultSize, wx.DefaultSize )
        
        bSizer1 = wx.BoxSizer( wx.VERTICAL )
        
        self.m_notebook1 = wx.Notebook( self, wx.ID_ANY, wx.DefaultPosition, wx.DefaultSize, 0 )
        self.m_panel1 = wx.Panel( self.m_notebook1, wx.ID_ANY, wx.DefaultPosition, wx.DefaultSize, wx.TAB_TRAVERSAL )
        self.m_panel1.SetToolTip( u"Basic controls, BoM selection and supported distributors." )
        
        bSizer3 = wx.BoxSizer( wx.VERTICAL )
        
        sbSizer2 = wx.StaticBoxSizer( wx.StaticBox( self.m_panel1, wx.ID_ANY, u"Files" ), wx.HORIZONTAL )
        
        m_comboBox_filesChoices = []
        self.m_comboBox_files = wx.ComboBox( sbSizer2.GetStaticBox(), wx.ID_ANY, u"Not selected files", wx.DefaultPosition, wx.DefaultSize, m_comboBox_filesChoices, 0 )
        self.m_comboBox_files.SetToolTip( u"BoM(s) file(s) to scrape.\nClick on the arrow to see/select one of the history files." )
        
        sbSizer2.Add( self.m_comboBox_files, 1, wx.ALL, 5 )
        
        self.m_button_openfile = wx.Button( sbSizer2.GetStaticBox(), wx.ID_ANY, u"Choose BoM", wx.DefaultPosition, wx.DefaultSize, 0 )
        self.m_button_openfile.SetToolTip( u"Click to choose the BoM(s) file(s)." )
        
        sbSizer2.Add( self.m_button_openfile, 0, wx.ALL, 5 )
        
        
        bSizer3.Add( sbSizer2, 0, wx.EXPAND|wx.TOP, 5 )
        
        bSizer4 = wx.BoxSizer( wx.HORIZONTAL )
        
        sbSizer3 = wx.StaticBoxSizer( wx.StaticBox( self.m_panel1, wx.ID_ANY, u"Distributors to scrape" ), wx.VERTICAL )
        
        m_checkList_distChoices = [wx.EmptyString]
        self.m_checkList_dist = wx.CheckListBox( sbSizer3.GetStaticBox(), wx.ID_ANY, wx.DefaultPosition, wx.DefaultSize, m_checkList_distChoices, 0 )
        self.m_checkList_dist.SetToolTip( u"Select the web distributor (or local) that will be used to scrape the prices.\nClick rigth to hot option." )
        
        sbSizer3.Add( self.m_checkList_dist, 1, wx.ALL|wx.EXPAND, 5 )
        
        
        bSizer4.Add( sbSizer3, 1, wx.EXPAND|wx.TOP|wx.LEFT, 5 )
        
        wSizer1 = wx.WrapSizer( wx.VERTICAL )
        
        bSizer6 = wx.BoxSizer( wx.VERTICAL )
        
        sbSizer31 = wx.StaticBoxSizer( wx.StaticBox( self.m_panel1, wx.ID_ANY, u"EDAs" ), wx.VERTICAL )
        
        m_listBox_edatoolChoices = []
        self.m_listBox_edatool = wx.ListBox( sbSizer31.GetStaticBox(), wx.ID_ANY, wx.DefaultPosition, wx.DefaultSize, m_listBox_edatoolChoices, 0 )
        self.m_listBox_edatool.SetToolTip( u"Choose the correct EDA software corresponding to the BoM file." )
        
        sbSizer31.Add( self.m_listBox_edatool, 1, wx.ALL|wx.EXPAND, 5 )
        
        self.m_checkBox_openXLS = wx.CheckBox( sbSizer31.GetStaticBox(), wx.ID_ANY, u"Open spreadsheet", wx.DefaultPosition, wx.DefaultSize, 0 )
        self.m_checkBox_openXLS.SetValue(True) 
        self.m_checkBox_openXLS.SetToolTip( u"Open the spreadsheet after finish the KiCost scrape." )
        
        sbSizer31.Add( self.m_checkBox_openXLS, 0, wx.ALL, 5 )
        
        
        bSizer6.Add( sbSizer31, 1, wx.TOP|wx.RIGHT|wx.EXPAND, 5 )
        
        self.m_button_run = wx.Button( self.m_panel1, wx.ID_ANY, u"KiCost it!", wx.DefaultPosition, wx.DefaultSize, 0 )
        self.m_button_run.SetToolTip( u"Click to run KiCost." )
        
        bSizer6.Add( self.m_button_run, 0, wx.ALL, 5 )
        
        
        wSizer1.Add( bSizer6, 1, wx.RIGHT|wx.EXPAND, 5 )
        
        
        bSizer4.Add( wSizer1, 1, wx.EXPAND, 5 )
        
        
        bSizer3.Add( bSizer4, 1, wx.EXPAND, 5 )
        
        fgSizer1 = wx.FlexGridSizer( 0, 4, 0, 0 )
        fgSizer1.SetFlexibleDirection( wx.BOTH )
        fgSizer1.SetNonFlexibleGrowMode( wx.FLEX_GROWMODE_SPECIFIED )
        
        self.m_gauge_process = wx.Gauge( self.m_panel1, wx.ID_ANY, 100, wx.DefaultPosition, wx.DefaultSize, wx.GA_HORIZONTAL )
        self.m_gauge_process.SetValue( 0 ) 
        self.m_gauge_process.SetToolTip( u"Percentage of the scrape process elapsed." )
        
        fgSizer1.Add( self.m_gauge_process, 1, wx.ALL|wx.EXPAND|wx.ALIGN_CENTER_HORIZONTAL|wx.ALIGN_CENTER_VERTICAL, 5 )
        
        self.m_staticText_progressInfo = wx.StaticText( self.m_panel1, wx.ID_ANY, wx.EmptyString, wx.DefaultPosition, wx.DefaultSize, 0 )
        self.m_staticText_progressInfo.Wrap( -1 )
        self.m_staticText_progressInfo.SetToolTip( u"Progress infromation." )
        
        fgSizer1.Add( self.m_staticText_progressInfo, 1, wx.ALL|wx.EXPAND|wx.ALIGN_CENTER_VERTICAL, 5 )
        
        
        bSizer3.Add( fgSizer1, 0, wx.EXPAND, 5 )
        
        self.m_staticText9 = wx.StaticText( self.m_panel1, wx.ID_ANY, u"Warnings, debug and error messages:", wx.DefaultPosition, wx.DefaultSize, 0 )
        self.m_staticText9.Wrap( -1 )
        bSizer3.Add( self.m_staticText9, 0, wx.ALL|wx.EXPAND, 5 )
        
        self.m_textCtrl_messages = wx.TextCtrl( self.m_panel1, wx.ID_ANY, wx.EmptyString, wx.DefaultPosition, wx.Size( -1,4 ), wx.HSCROLL|wx.TE_MULTILINE|wx.TE_READONLY )
        self.m_textCtrl_messages.SetToolTip( u"Process messages and warnings.\nClick right to copy or save the log." )
        self.m_textCtrl_messages.SetMinSize( wx.Size( -1,4 ) )
        
        bSizer3.Add( self.m_textCtrl_messages, 1, wx.ALL|wx.EXPAND|wx.ALIGN_CENTER_HORIZONTAL, 5 )
        
        
        self.m_panel1.SetSizer( bSizer3 )
        self.m_panel1.Layout()
        bSizer3.Fit( self.m_panel1 )
        self.m_notebook1.AddPage( self.m_panel1, u"BoM", True )
        self.m_panel2 = wx.Panel( self.m_notebook1, wx.ID_ANY, wx.DefaultPosition, wx.DefaultSize, wx.TAB_TRAVERSAL )
        self.m_panel2.SetToolTip( u"KiCost general configurations tab." )
        
        bSizer8 = wx.BoxSizer( wx.VERTICAL )
        
        wSizer2 = wx.WrapSizer( wx.HORIZONTAL )
        
        bSizer9 = wx.BoxSizer( wx.VERTICAL )
        
        self.m_staticText2 = wx.StaticText( self.m_panel2, wx.ID_ANY, u"Parallel process", wx.DefaultPosition, wx.DefaultSize, 0 )
        self.m_staticText2.Wrap( -1 )
        bSizer9.Add( self.m_staticText2, 0, wx.ALL, 5 )
        
        self.m_spinCtrl_np = wx.SpinCtrl( self.m_panel2, wx.ID_ANY, wx.EmptyString, wx.DefaultPosition, wx.DefaultSize, wx.SP_ARROW_KEYS, 1, 30, 6 )
        self.m_spinCtrl_np.SetToolTip( u"Set the number of parallel processes used for web scraping the parts data." )
        
        bSizer9.Add( self.m_spinCtrl_np, 0, wx.ALL, 5 )
        
        self.m_staticText3 = wx.StaticText( self.m_panel2, wx.ID_ANY, u"Scrap retries", wx.DefaultPosition, wx.DefaultSize, 0 )
        self.m_staticText3.Wrap( -1 )
        bSizer9.Add( self.m_staticText3, 0, wx.ALL, 5 )
        
        self.m_spinCtrl_retries = wx.SpinCtrl( self.m_panel2, wx.ID_ANY, wx.EmptyString, wx.DefaultPosition, wx.DefaultSize, wx.SP_ARROW_KEYS, 4, 200, 0 )
        self.m_spinCtrl_retries.SetToolTip( u"Specify the number of attempts to retrieve part data from a website." )
        
        bSizer9.Add( self.m_spinCtrl_retries, 0, wx.ALL, 5 )
        
        self.m_staticText7 = wx.StaticText( self.m_panel2, wx.ID_ANY, u"Throttling delay (s)", wx.DefaultPosition, wx.DefaultSize, 0 )
        self.m_staticText7.Wrap( -1 )
        bSizer9.Add( self.m_staticText7, 0, wx.ALL, 5 )
        
        self.m_spinCtrlDouble_throttling = wx.SpinCtrlDouble( self.m_panel2, wx.ID_ANY, wx.EmptyString, wx.DefaultPosition, wx.DefaultSize, wx.SP_ARROW_KEYS, 0, 5, 0, 0.1 )
        self.m_spinCtrlDouble_throttling.SetToolTip( u"Specify minimum delay (in seconds) between successive accesses to a distributor's website.\nUsed when the websites not accept successive accesses." )
        
        bSizer9.Add( self.m_spinCtrlDouble_throttling, 0, wx.ALL, 5 )
        
        
        wSizer2.Add( bSizer9, 1, wx.TOP|wx.LEFT, 5 )
        
        bSizer11 = wx.BoxSizer( wx.VERTICAL )
        
        self.m_checkBox_collapseRefs = wx.CheckBox( self.m_panel2, wx.ID_ANY, u"Collapse refs", wx.DefaultPosition, wx.DefaultSize, 0 )
        self.m_checkBox_collapseRefs.SetValue(True) 
        self.m_checkBox_collapseRefs.SetToolTip( u"Collapse the references in the spreadsheet.\n'R1,R2,R3,R4,R9' become 'R1-R4,R9' with checked." )
        
        bSizer11.Add( self.m_checkBox_collapseRefs, 0, wx.ALL, 5 )
        
        self.m_staticText8 = wx.StaticText( self.m_panel2, wx.ID_ANY, u"Debug level", wx.DefaultPosition, wx.DefaultSize, 0 )
        self.m_staticText8.Wrap( -1 )
        bSizer11.Add( self.m_staticText8, 0, wx.ALL, 5 )
        
        self.m_spinCtrl_debugLvl = wx.SpinCtrl( self.m_panel2, wx.ID_ANY, wx.EmptyString, wx.DefaultPosition, wx.DefaultSize, wx.SP_ARROW_KEYS, 0, 10, 0 )
        bSizer11.Add( self.m_spinCtrl_debugLvl, 0, wx.ALL, 5 )
        
        self.m_checkBox_quite = wx.CheckBox( self.m_panel2, wx.ID_ANY, u"Quiet mode", wx.DefaultPosition, wx.DefaultSize, 0 )
        self.m_checkBox_quite.SetToolTip( u"Enable quiet mode with no warnings." )
        
        bSizer11.Add( self.m_checkBox_quite, 0, wx.ALL, 5 )
        
        self.m_checkBox_overwrite = wx.CheckBox( self.m_panel2, wx.ID_ANY, u"Overwrite file", wx.DefaultPosition, wx.DefaultSize, 0 )
        self.m_checkBox_overwrite.SetValue(True) 
        self.m_checkBox_overwrite.SetToolTip( u"Allow overwriting of an existing spreadsheet." )
        
        bSizer11.Add( self.m_checkBox_overwrite, 0, wx.ALL, 5 )
        
        
        wSizer2.Add( bSizer11, 1, wx.TOP|wx.RIGHT, 5 )
        
        
        bSizer8.Add( wSizer2, 1, wx.ALL|wx.EXPAND|wx.ALIGN_CENTER_HORIZONTAL, 5 )
        
        self.m_staticText4 = wx.StaticText( self.m_panel2, wx.ID_ANY, u"Extra commands", wx.DefaultPosition, wx.DefaultSize, 0 )
        self.m_staticText4.Wrap( -1 )
        bSizer8.Add( self.m_staticText4, 0, wx.ALL, 5 )
        
        self.m_textCtrlextracmd = wx.TextCtrl( self.m_panel2, wx.ID_ANY, wx.EmptyString, wx.DefaultPosition, wx.DefaultSize, 0 )
        self.m_textCtrlextracmd.SetToolTip( u" Here use the kicost extra commands. In the terminal/command type`kicost --help` to check the list." )
        
        bSizer8.Add( self.m_textCtrlextracmd, 0, wx.ALL|wx.EXPAND, 5 )
        
        
        self.m_panel2.SetSizer( bSizer8 )
        self.m_panel2.Layout()
        bSizer8.Fit( self.m_panel2 )
        self.m_notebook1.AddPage( self.m_panel2, u"KiCost config", False )
        self.m_panel3 = wx.Panel( self.m_notebook1, wx.ID_ANY, wx.DefaultPosition, wx.DefaultSize, wx.TAB_TRAVERSAL )
        self.m_panel3.SetToolTip( u"About the software, version installation and update found." )
        
        bSizer2 = wx.BoxSizer( wx.VERTICAL )
        
        bSizer10 = wx.BoxSizer( wx.HORIZONTAL )
        
        self.m_bitmap_icon = wx.StaticBitmap( self.m_panel3, wx.ID_ANY, wx.NullBitmap, wx.DefaultPosition, wx.DefaultSize, 0 )
        bSizer10.Add( self.m_bitmap_icon, 0, wx.ALL, 5 )
        
        bSizer111 = wx.BoxSizer( wx.VERTICAL )
        
        self.m_staticText_version = wx.StaticText( self.m_panel3, wx.ID_ANY, u"MyLabel", wx.DefaultPosition, wx.DefaultSize, 0 )
        self.m_staticText_version.Wrap( -1 )
        bSizer111.Add( self.m_staticText_version, 1, wx.ALL, 5 )
        
        self.m_staticText_update = wx.StaticText( self.m_panel3, wx.ID_ANY, u"Update info", wx.DefaultPosition, wx.DefaultSize, 0 )
        self.m_staticText_update.Wrap( -1 )
        bSizer111.Add( self.m_staticText_update, 0, wx.ALL, 5 )
        
        
        bSizer10.Add( bSizer111, 1, wx.EXPAND, 5 )
        
        
        bSizer2.Add( bSizer10, 0, wx.ALL|wx.EXPAND, 5 )
        
        self.m_staticText_credits = wx.StaticText( self.m_panel3, wx.ID_ANY, u"MyLabel", wx.DefaultPosition, wx.DefaultSize, 0 )
        self.m_staticText_credits.Wrap( -1 )
        bSizer2.Add( self.m_staticText_credits, 1, wx.ALL|wx.ALIGN_CENTER_HORIZONTAL|wx.EXPAND, 5 )
        
        
        self.m_panel3.SetSizer( bSizer2 )
        self.m_panel3.Layout()
        bSizer2.Fit( self.m_panel3 )
        self.m_notebook1.AddPage( self.m_panel3, u"About", False )
        
        bSizer1.Add( self.m_notebook1, 1, wx.EXPAND |wx.ALL, 5 )
        
        
        self.SetSizer( bSizer1 )
        self.Layout()
        
        self.Centre( wx.BOTH )
        
        # Connect Events
        self.Bind( wx.EVT_CLOSE, self.app_close )
        self.m_notebook1.Bind( wx.EVT_NOTEBOOK_PAGE_CHANGED, self.wxPanel_change )
        self.m_comboBox_files.Bind( wx.EVT_COMBOBOX, self.m_comboBox_files_selecthist )
        self.m_button_openfile.Bind( wx.EVT_BUTTON, self.button_openfile )
        self.m_checkList_dist.Bind( wx.EVT_RIGHT_DOWN, self.m_textCtrl_distributors_rClick )
        self.m_button_run.Bind( wx.EVT_BUTTON, self.button_run )
        self.m_textCtrl_messages.Bind( wx.EVT_RIGHT_DOWN, self.m_textCtrl_messages_rClick )
        self.m_bitmap_icon.Bind( wx.EVT_LEFT_DOWN, self.m_bitmap_icon_click )
        
        #### **  End of the guide code generated by wxFormBulilder software, available in <https://github.com/wxFormBuilder/wxFormBuilder/>  ** ####
        
        self.updateChecked = False 
        self.set_properties()
        self.SetDropTarget( FileDropTarget(self) ) # Start the drop file in all the window.
        self.m_textCtrl_messages.AppendText( 'Loaded KiCost v.' + __version__ )
    
    def __del__( self ):
        pass

    # Virtual event handlers, overide them in your derived class

    #----------------------------------------------------------------------
    def app_close( self, event ):
        ''' @brief Close event, used to call the save settings.'''
        event.Skip()
        self.save_properties()

    #----------------------------------------------------------------------
    def m_bitmap_icon_click( self, event ):
        ''' @brief Open the official software web page in the default browser.'''
        event.Skip()
        webbrowser.open(PAGE_OFFICIAL)

    #----------------------------------------------------------------------
    def wxPanel_change( self, event ):
        ''' @brief Check version to update if the "About" tab.'''
        event.Skip()
        if event.GetSelection()==2: # Is the last page (about page).
            
            def checkUpdate():
                '''Check for updates.'''
                try:
                    updateChecked = self.updateChecked
                except:
                    updateChecked = False
                if not updateChecked:
                    self.m_staticText_update.SetLabel('Checking by updates...')
                    try:
                        html = fake_browser(PAGE_UPDATE)
                        offical_last_version = re.findall('kicost (\d+\.\d+\.\d+)', str(html), flags=re.IGNORECASE)[0]
                        if StrictVersion(offical_last_version) > StrictVersion(__version__):
                            self.m_staticText_update.SetLabel('New version (v.'
                                + offical_last_version
                                + ') founded.\nClick <here> to update.')
                            self.m_staticText_update.Bind( wx.EVT_LEFT_UP, self.m_staticText_update_click )
                        else:
                            self.m_staticText_update.SetLabel('KiCost is up to date.')
                    except:
                        self.m_staticText_update.SetLabel('Update information not founded.')
                    finally:
                        self.updateChecked = True
            
            wx.CallLater(50, checkUpdate) # Thread optimized for graphical elements change.

    #----------------------------------------------------------------------
    def m_textCtrl_messages_rClick( self, event ):
        ''' @brief Open the context menu with save log options.'''
        event.Skip()
        self.PopupMenu(menuMessages(self), event.GetPosition())

    #----------------------------------------------------------------------
    def m_textCtrl_distributors_rClick( self, event ):
        ''' @brief Open the context menu with distributors options.'''
        event.Skip()
        self.PopupMenu(menuDistributors(self), event.GetPosition())

    #----------------------------------------------------------------------
    def m_comboBox_files_selecthist( self, event):
        ''' @brief Update the select EDA module tool when changed the file selected in the history, update the order and delete if file not existent.'''
        event.Skip()
        # Check if the file in the file name exist and
        # update the history sequence, if don't, remove it.
        histSelected = event.GetSelection()
        fileNames = event.GetString()
        self.m_comboBox_files.Delete(histSelected)
        if all(os.path.isfile(f) for f in re.split(SEP_FILES, fileNames) ):
            self.m_comboBox_files.Insert( fileNames, 0 )
            self.updateEDAselection() # Auto-select the EDA module.
        else:
            self.m_comboBox_files.SetValue('')

    #----------------------------------------------------------------------
    def updateEDAselection( self ):
        ''' @brief Update the EDA selection in the listBox based on the comboBox actual text.'''
        fileNames = re.split(SEP_FILES, self.m_comboBox_files.GetValue())
        if len(fileNames)==1:
            eda_module = file_eda_match(fileNames[0])
            if eda_module:
                self.m_listBox_edatool.SetSelection( self.m_listBox_edatool.FindString(eda_tool_dict[eda_module]['label']) )
        elif len(fileNames)>1:
            # Check if all the EDA are the same. For different ones,
            # the guide is not able now to deal, need improvement
            # on `self.m_listBox_edatool`.
            eda_module = file_eda_match(fileNames[0])
            for fName in fileNames[1:]:
                if file_eda_match(fName) != eda_module:
                    return
            if eda_module:
                self.m_listBox_edatool.SetSelection( self.m_listBox_edatool.FindString(eda_tool_dict[eda_module]['label']) )

    #----------------------------------------------------------------------
    def m_staticText_update_click( self, event ):
        ''' @brief Open the page to download the last version.'''
        event.Skip()
        #print('Download the update and install -- missing, running manually')
        webbrowser.open(PAGE_UPDATE)

    #----------------------------------------------------------------------
    def button_openfile( self, event ):
        """ @brief Create and show the Open FileDialog """
        event.Skip()
        
        actualDir = (os.getcwd() if self.m_comboBox_files.GetValue() else \
            os.path.dirname(os.path.abspath( self.m_comboBox_files.GetValue() )) )
        
        dlg = wx.FileDialog(
            self, message = "Select BoM(s)",
            defaultDir = actualDir, 
            defaultFile = "",
            wildcard = WILDCARD,
            style = wx.FD_OPEN | wx.FD_MULTIPLE | wx.FD_CHANGE_DIR
            )
        if dlg.ShowModal() == wx.ID_OK:
            paths = dlg.GetPaths()
            self.addFile( paths )
        dlg.Destroy()

    #----------------------------------------------------------------------
    def addFile( self, filesName ):
        ''' @brief Add the file(s) to the history, updating it (and delete the too old).'''
        fileBOM = SEP_FILES.join( sorted(filesName, key=str.lower) ) # Add the files sorted.
        if self.m_comboBox_files.FindString(fileBOM)==wx.NOT_FOUND:
            self.m_comboBox_files.Insert( fileBOM, 0 )
        self.m_comboBox_files.SetValue( fileBOM )
        try:
            self.m_comboBox_files.Delete(FILE_HIST_QTY-1) # Keep 10 files on history.
        except:
            pass
        self.updateEDAselection()

    #----------------------------------------------------------------------
    def button_run( self, event ):
        ''' @brief Call to run KiCost.'''
        event.Skip()
        def run_kicost_guide():
            class EtaStream(object):
                def __init__(self, widget):
                    #super(self.__class__, self).__init__()
                    pass
                def write(self, bar):
                    sys.stderr.write(bar)
                    try:
                        #Progress:   3%|█                               | 3/90 [00:13<19:42, 13.59s/part]
                        print('-------', bar)
                        #perc = re.findall('(\d)\%\|')[0]
                        #desc = re.findall('\| (.+)$')[0]
                        #print('---', perc, desc)#TODO
                        #self.m_gauge_process.SetValue(perc)
                        #self.m_staticText_progressInfo.SetLabel(desc)
                    except:
                        pass
                    def flush(self):
                        sys.stderr.flush()
            class GUILoggingHandler(object):
                def __init__(self, widget):
                    #super(self.__class__, self).__init__()
                    self.widget = widget
                def write(self, msg):
                    try:
                        self.widget.AppendText( msg )
                    except:
                        sys.__stdout__(msg)
                def flush(self):#TODO
                    sys.stderr.flush()
            self.m_button_openfile.Enable( False )
            #sys.stdout = GUILoggingHandler(self.m_textCtrl_messages)
            #sys.errout = GUILoggingHandler(self.m_textCtrl_messages)
            
            self.save_properties() # Save the current graphical configuration before call the KiCost motor.
            self.run() # Run KiCost.
            
            # Restore the channel print output to terminal.
            sys.stdout = sys.__stdout__
            sys.stderr = sys.__stderr__
            
            self.m_button_openfile.Enable( True )
        def run_kicost_guide2():
            '''Run the as a Thread out of the box wxPython'''
            self.m_gauge_process.SetValue(0)
            kicost_motor_thread = threading.Thread(target=run_kicost_guide)
            kicost_motor_thread.start()
        wx.CallLater(10, run_kicost_guide2) # Necessary to not '(core dumped)' with wxPython.

    #----------------------------------------------------------------------
    def run( self ):
        ''' @brief Run KiCost.
            Run KiCost in the GUI interface updating the process bar and messages.'''
        #self.m_gauge_process.SetValue(0)
        #self.m_textCtrl_messages.Clear() # Clear the messages to appear just the last run.
        
        class argments:
            pass
        args = argments()
        
        args.input = re.split(SEP_FILES, self.m_comboBox_files.GetValue())
        for f in args.input:
            if not os.path.isfile(f):
                print('No valid file(s) selected.')
                return # Not a valid file(s).
        
        spreadsheet_file = re.split(SEP_FILES, self.m_comboBox_files.GetValue())
        if len(spreadsheet_file)==1:
            spreadsheet_file = os.path.splitext( spreadsheet_file[0] )[0] + '.xlsx'
        else:
            spreadsheet_file = output_filename_multipleinputs( spreadsheet_file )
        # Handle case where output is going into an existing spreadsheet file.
        if os.path.isfile(spreadsheet_file):
            if not self.m_checkBox_overwrite.GetValue():
                dlg = wx.MessageDialog(self, 
                    "The file output \'{}\' already exit, do you wnat overwrite?".format(
                                os.path.basename(spreadsheet_file)
                            ),
                    "Confirm Overwrite", wx.YES_NO|wx.YES_DEFAULT|wx.ICON_QUESTION|wx.STAY_ON_TOP|wx.CENTER)
                result = dlg.ShowModal()
                dlg.Destroy()
                if result==wx.ID_NO:
                    print('Not able to overwrite \'{}\'...'.format(
                                os.path.basename(spreadsheet_file)
                            )
                        )
                    return
        args.output = spreadsheet_file
        
        if self.m_textCtrlextracmd.GetValue():
            extra_commands = ' ' + self.m_textCtrlextracmd.GetValue()
        else:
            extra_commands = []
        
        def str_to_arg(commands):
            try:
                for c in commands:
                    try:
                        return ''.join( re.findall(c+' (.+)', extra_commands))
                    except:
                        continue
            except:
                pass
            finally:
                return ''
        args.fields = str_to_arg(['--fields', '-f']).split()
        args.ignore_fields = str_to_arg(['--ignore_fields', '-ign']).split()
        args.group_fields = str_to_arg(['--group_fields', '-grp']).split()
        args.variant = str_to_arg(['--variant', '-var'])
        
        num_processes = self.m_spinCtrl_np.GetValue() # Parallels process scrapping.
        args.retries = self.m_spinCtrl_retries.GetValue() # Retry time in the scraps.
        args.throttling_delay = self.m_spinCtrlDouble_throttling.GetValue() # Delay between consecutive scrapes.
        args.collapse_refs = self.m_checkBox_collapseRefs.GetValue() # Collapse refs in the spreadsheet.
        
        if self.m_listBox_edatool.GetStringSelection():
            for k,v in eda_tool_dict.items():
               if v['label']==self.m_listBox_edatool.GetStringSelection():
                  eda_module = v['module']
                  break
        args.eda_tool = eda_module
        
        # Get the current distributors to scrape.
        choisen_dist = list(self.m_checkList_dist.GetCheckedItems())
        if choisen_dist:
            dist_list = []
            #choisen_dist = [self.m_checkList_dist.GetString(idx) for idx in choisen_dist]
            for idx in choisen_dist:
                label = self.m_checkList_dist.GetString(idx)
                for k,v in distributor_dict.items():
                    if v['label']==label:
                        dist_list.append( v['module'] )
                        break
        else:
            dist_list = None
        args.include = dist_list
        
        # Run KiCost main function and print in the log the elapsed time.
        start_time = time.time()
        try:
            kicost(in_file=args.input, eda_tool_name=args.eda_tool,
                out_filename=args.output, collapse_refs=args.collapse_refs,
                user_fields=args.fields, ignore_fields=args.ignore_fields,
                group_fields=args.group_fields, variant=args.variant,
                dist_list=args.include, num_processes=num_processes,
                scrape_retries=args.retries, throttling_delay=args.throttling_delay)
            if self.m_checkBox_openXLS.GetValue():
                print('Opening the output file \'{}\'...'.format(
                                    os.path.basename(spreadsheet_file)
                                )
                            )
                open_file(spreadsheet_file)
        except Exception as e:
            print(e)
        print('Elapsed time: {} seconds'.format(time.time() - start_time) )
        #self.m_gauge_process.SetValue(100)
        
        return

    #----------------------------------------------------------------------
    def runTerminal( self ):
        ''' @brief Run KiCost in CLI interface using the GUI settings.'''
        
        # Get the current distributors to scrape.
        choisen_dist = list(self.m_checkList_dist.GetCheckedItems())
        if choisen_dist:
            dist_list = ' --include'
            #choisen_dist = [self.m_checkList_dist.GetString(idx) for idx in choisen_dist]
            for idx in choisen_dist:
                label = self.m_checkList_dist.GetString(idx)
                for k,v in distributor_dict.items():
                    if v['label']==label:
                        dist_list += ' ' + v['module']
                        break
            #choisen_dist = ' --include ' + ' '.join(choisen_dist)
        else:
            dist_list = ''
        
        command = ("kicost"
            + " --input " + ' '.join(['"'+fileN+'"' for fileN in re.split(SEP_FILES, self.m_comboBox_files.GetValue())])
            + " --num_processes " + str(self.m_spinCtrl_np.GetValue()) # Parallels process scrapping.
            + " --retries " + str(self.m_spinCtrl_retries.GetValue()) # Retry time in the scraps.
            + " --throttling " + str(self.m_spinCtrlDouble_throttling.GetValue()) # Delay between consecutive scrapes.
            + " --overwrite" * self.m_checkBox_overwrite.GetValue()
            + " --no_collpase" * ( not self.m_checkBox_collapseRefs.GetValue() )
            + (" --debug " + str(self.m_spinCtrl_debugLvl.GetValue()) if self.m_spinCtrl_debugLvl.GetValue() > 0 else "") # Degub level opiton.
            + " --quiet" * self.m_checkBox_quite.GetValue()
            + dist_list
            )
        
        if self.m_listBox_edatool.GetStringSelection():
            for k,v in eda_tool_dict.items():
               if v['label']==self.m_listBox_edatool.GetStringSelection():
                  eda_module = v['module']
                  break
            command += " -eda " + eda_module
        
        if self.m_textCtrlextracmd.GetValue():
            command += ' ' + self.m_textCtrlextracmd.GetValue()
        
        if self.m_checkBox_openXLS.GetValue():
            spreadsheet_file = re.split(SEP_FILES, self.m_comboBox_files.GetValue())
            if len(spreadsheet_file)==1:
                spreadsheet_file = os.path.splitext( spreadsheet_file[0] )[0] + '.xlsx'
            else:
                spreadsheet_file = output_filename_multipleinputs( spreadsheet_file )
            
            if platform.system()=='Linux':
                command += ' && xdg-open ' + '"' + spreadsheet_file + '"'
            elif platform.system()=='Windows':
                command += ' && explorer ' + '"' + spreadsheet_file + '"'
            elif platform.system()=='Darwin': # Mac-OS
                command += ' && open -n ' + '"' + spreadsheet_file + '"'
            else:
                print('Not recognized OS.')
        
        command += '&' # Run as other process.
        print("Running: ", command)
        subprocess.call(command, shell=True) # `os.system`not accept the "&&"

    #----------------------------------------------------------------------
    def set_properties(self):
        ''' @brief Set the current proprieties of the graphical elements.'''
        actualDir = os.path.dirname(os.path.abspath(__file__)) # Application dir.
        
        # Set the aplication windows title and configurations
        self.SetTitle('KiCost v.' + __version__)
        self.SetIcon(wx.Icon(actualDir + os.sep + 'kicost.ico', wx.BITMAP_TYPE_ICO))
        
        # Current distrubutors module recognized.
        distributors_list = sorted( [ distributor_dict[d]['label'] for d in distributor_dict.keys() ] )
        self.m_checkList_dist.Clear()
        self.m_checkList_dist.Append(distributors_list)
        for idx in range(len(distributors_list)):
            self.m_checkList_dist.Check(idx,True) # All start checked (after is modifed by the configuration file).
        
        # Current EDA tools module recognized.
        eda_names = sorted( [ eda_tool_dict[eda]['label'] for eda in eda_tool_dict.keys() ] )
        self.m_listBox_edatool.Clear()
        self.m_listBox_edatool.Append(eda_names)
        
        # Credits and other informations, search by `AUTHOR.rst` file.
        self.m_staticText_version.SetLabel( 'KiCost version ' + __version__ )
        self.m_bitmap_icon.SetIcon(wx.Icon(actualDir + os.sep + 'kicost.ico', wx.BITMAP_TYPE_ICO))
        try:
            credits_file = open(actualDir + os.sep+'..'+os.sep + 'kicost-' + __version__ + '.egg-info' + os.sep + 'AUTHOR.rst')
            credits = credits_file.read()
            credits_file.close()
        except:
            credits = r'''=======
            Credits
            =======
            Development Lead
            ----------------
            * XESS Corporation <info@xess.com>
            Contributors
            ------------
            * Oliver Martin: https://github.com/oliviermartin
            * Timo Alho: https://github.com/timoalho
            * Steven Johnson: https://github.com/stevenj
            * Diorcet Yann: https://github.com/diorcety
            * Giacinto Luigi Cerone https://github.com/glcerone
            * Hildo Guillardi Júnior https://github.com/hildogjr
            * Adam Heinrich https://github.com/adamheinrich
            ------------
            GUI by Hildo Guillardi Júnior
            '''
            credits = re.sub(r'\n[\t ]+', '\n', credits)  # Remove leading whitespace
        self.m_staticText_credits.SetLabel(credits)
        
        # Recovery the last configurations used (found the folder of the file by the OS).
        self.restore_properties()
        
        # Files in the history.
        #if not self.m_comboBox_files.IsListEmpty(): # If have some history, set to the last used file.
        #    self.m_comboBox_files.IsListEmpty(0)

    #----------------------------------------------------------------------
    def restore_properties(self):
        ''' @brief Restore the current proprieties of the graphical elements.'''
        try:
            configHandle = wx.Config(CONFIG_FILE)
            
            def str_to_wxpoint(s):
                '''Convert a string tuple into a  wxPoint'''
                p = re.findall('\d+', s)
                return wx.Point(int(p[0]), int(p[1]))
            def str_to_wxsize(s):
                '''Convert a string tuple into a  wxRect'''
                p = re.findall('\d+', s)
                return wx.Size(int(p[0]), int(p[1]))
            
            entryCount = 0
            while True:
                entry = configHandle.GetNextEntry(entryCount)
                if not entry[0]:
                    break
                entryCount+=1 #Count the entry numbers and go to next one in next iteration.
                entry = entry[1]
                entry_value = configHandle.Read(entry)
                
                # Resize and reposition the window frame.
                if entry==GUI_POSITION_ENTRY:
                    self.SetPosition(str_to_wxpoint(entry_value))
                    continue
                elif entry==GUI_SIZE_ENTRY:
                    self.SetSize(str_to_wxsize(entry_value))
                    continue
                
                try:
                    # Find the wxPython element handle to access the methods.
                    wxElement_handle = self.__dict__[entry]
                    # Each wxPython object have a specific parameter value
                    # to be saved and restored in the software initialization.
                    if isinstance(wxElement_handle, wx._core.TextCtrl):
                        wxElement_handle.SetValue( entry_value )
                    elif isinstance(wxElement_handle, wx._core.CheckBox):
                        wxElement_handle.SetValue( (True if entry_value=='True' else False) )
                    elif isinstance(wxElement_handle, wx._core.CheckListBox):
                        value = re.split(',', entry_value )
                        for idx in range(wxElement_handle.GetCount()): # Reset all checked.
                            wxElement_handle.Check(idx, False)
                        for dist_checked in value: # Check only the founded names.
                            idx = wxElement_handle.FindString( dist_checked )
                            if idx!=wx.NOT_FOUND:
                                wxElement_handle.Check(idx, True)
                    elif isinstance(wxElement_handle, wx._core.SpinCtrl):
                        wxElement_handle.SetValue( int(entry_value) )
                    elif isinstance(wxElement_handle, wx._core.SpinCtrlDouble):
                        wxElement_handle.SetValue( float(entry_value) )
                    elif isinstance(wxElement_handle, wx._core.ComboBox):
                        value = re.split(',', entry_value)
                        for element in value:
                            if element:
                                wxElement_handle.Append( element )
                    elif isinstance(wxElement_handle, wx._core.ListBox):
                        wxElement_handle.SetSelection( wxElement_handle.FindString( entry_value ) )
                    elif isinstance(wxElement_handle, wx._core.Notebook):
                        wxElement_handle.SetSelection( int(entry_value) )
                    # Others wxWidgets graphical elements with not saved configurations.
                    #elif isinstance(wxElement_handle, wx._core.):
                    #elif isinstance(wxElement_handle, wx._core.):configHandle
                    #elif isinstance(wxElement_handle, wx._core.StaticBitmap):
                    #elif isinstance(wxElement_handle, wx._core.Panel):
                    #elif isinstance(wxElement_handle, wx._core.Button):
                    #elif isinstance(wxElement_handle, wx._core.StaticText):
                except KeyError:
                    continue
                
            del configHandle # Close the file / Windows registry sock.
        except Exception as e:
            self.m_textCtrl_messages.AppendText('Configurations not recovered: <'+str(e)+'>.')

    #----------------------------------------------------------------------
    def save_properties(self):
        ''' @brief Save the current proprieties of the graphical elements.'''
        try:
            configHandle = wx.Config(CONFIG_FILE)
            
            # Save position and size.
            configHandle.Write(GUI_POSITION_ENTRY, str(self.GetPosition()))
            configHandle.Write(GUI_SIZE_ENTRY, str(self.GetSize()))
            
            # Sweep all elements in `self()` to find the grafical ones
            # instance of the wxPython and salve the specific configuration.
            for wxElement_name, wxElement_handle in self.__dict__.items():
                try:
                    # Each wxPython object have a specific parameter value
                    # to be saved and restored in the software initialization.
                    if isinstance(wxElement_handle, wx._core.TextCtrl) and wxElement_name != 'm_textCtrl_messages':
                        # Save each TextCtrl (TextBox) that is not the status messages.
                        configHandle.Write(wxElement_name, wxElement_handle.GetValue() )
                    elif isinstance(wxElement_handle, wx._core.CheckBox):
                        configHandle.Write(wxElement_name, ('True' if wxElement_handle.GetValue() else 'False') )
                    elif isinstance(wxElement_handle, wx._core.CheckListBox):
                        value = [wxElement_handle.GetString(idx) for idx in wxElement_handle.GetCheckedItems()]
                        configHandle.Write(wxElement_name, ','.join(value) )
                    elif isinstance(wxElement_handle, wx._core.SpinCtrl) or isinstance(wxElement_handle, wx._core.SpinCtrlDouble):
                        configHandle.Write(wxElement_name, str(wxElement_handle.GetValue()) )
                    elif isinstance(wxElement_handle, wx._core.ComboBox):
                        value = [wxElement_handle.GetString(idx) for idx in range(wxElement_handle.GetCount())]
                        configHandle.Write(wxElement_name, ','.join(value) )
                    elif isinstance(wxElement_handle, wx._core.ListBox):
                        configHandle.Write(wxElement_name, wxElement_handle.GetStringSelection() )
                    elif isinstance(wxElement_handle, wx._core.Notebook):
                        configHandle.Write(wxElement_name, str(wxElement_handle.GetSelection()) )
                    # Others wxWidgets graphical elements with not saved configurations.
                    #elif isinstance(wxElement_handle, wx._core.):configHandle
                    #elif isinstance(wxElement_handle, wx._core.StaticBitmap):
                    #elif isinstance(wxElement_handle, wx._core.Panel):
                    #elif isinstance(wxElement_handle, wx._core.Button):
                    #elif isinstance(wxElement_handle, wx._core.StaticText):
                except KeyError:
                    continue
            
            del configHandle # Close the file / Windows registry sock.
        except Exception as e:
            self.m_textCtrl_messages.AppendText('Configurations not saved: <'+str(e)+'>.')





#######################################################################

def kicost_gui():
    ''' @brief Load the graphical interface.'''
    app = wx.App(redirect=False)
    frame = formKiCost(None)
    frame.Show()
    app.MainLoop()

def kicost_gui_run(fileName):
    ''' @brief Execute the `fileName`under KiCost loading the graphical interface.
        @param LIst of the file name.
    '''
    app = wx.App(redirect=False)
    frame = formKiCost(None)
#    frame.Show()
    frame.m_comboBox_files.SetValue('"' + '", "'.join(fileName) + '"')
    frame.updateEDAselection()
    frame.runTerminal()
#    app.MainLoop()
